'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import numpy as np
import copy
from nncodec.nnc_core import common
from nncodec.extensions import deepCABAC
from nncodec.nnc_core.coder import hls
from nncodec.nnc_core.nnr_model import NNRModelAccess, W_TYPES
from nncodec.nnc_core.hdsp.hdsp_tool import HDSP_OPTS_OFF


def derive_sorted_codebook_from_tensor(tensor):
    originalShape = tensor.shape
    codebook, indices = np.unique(tensor, return_inverse=True)
    reshaped_indices = indices.reshape( originalShape )
    return codebook, reshaped_indices.astype('int32')

def get_codebook_offset( codebook, indices, cabac_unary_length_minus1 ):
    codebookOffset = 0
    if indices.dtype == np.int32:
        codebookOffset = -1
        minBits = None
        for cb in range( len( codebook ) ):
            encoder = deepCABAC.Encoder()
            encoder.initCtxModels( cabac_unary_length_minus1, 1 )
            indexes = indices - cb
            hdsp_opts = HDSP_OPTS_OFF()
            encoder.encodeLayer(indexes, 0, 0, 0, 0, 0, np.zeros(indexes.shape[0], dtype=np.int32), *hdsp_opts, 0, 0)
            bits = len( encoder.finish().tobytes() )
            if minBits == None or bits < minBits:
                minBits = bits
                codebookOffset = cb

    indexes = indices - codebookOffset

    return codebook, indexes, codebookOffset

def get_best_egk(codebook, codebookOffset):
    cb_hls = {}
    cb_hls["CbZeroOffset__"] = codebookOffset
    cb_hls["codebook_size__"] = len( codebook )
    cb_hls["codebook__"] = codebook

    min_bytes_cb = None
    for i in range(16):
        cb_hls["codebook_egk__"] = i
        bs = bytearray()
        w = hls.BitWriter(bs)
        hls_enc = hls.Coder( w, cb_hls )
        hls_enc.codebook("")
        bytes_cb = w.getNumBitsTouched()
        if min_bytes_cb is None or bytes_cb < min_bytes_cb:
            min_bytes_cb = bytes_cb
            best_egk = i
    return best_egk, min_bytes_cb

def get_codebook_bytes(codebook, codebookOffset, cbEgk):
    cb_hls = {}
    cb_hls["CbZeroOffset__"] = codebookOffset
    cb_hls["codebook_size__"] = len( codebook )
    cb_hls["codebook__"] = codebook
    cb_hls["codebook_egk__"] = cbEgk
    
    bs = bytearray()
    w = hls.BitWriter(bs)
    hls_enc = hls.Coder( w, cb_hls )
    hls_enc.codebook("")
    bytes_cb = w.getNumBitsTouched()
   
    return bytes_cb

def check_array_all_zero_or_scalar(x):
    if np.isscalar(x):
        return isinstance(x, (int, np.integer))
    elif isinstance(x, np.ndarray) and x.ndim == 0:
        return isinstance(x.item(), (int, np.integer))
    elif isinstance(x, np.ndarray) and np.all(x == 0):
        return True
    else:
        return False

def clip_for_int_aligned(quantizedValues, approx_info, approx_data_in, model_info, param):
    bw_ap = approx_info["integer_aligned_bitdepth"]
    if bw_ap > 7 or model_info["parameter_type"][param] in W_TYPES:
        bw = bw_ap
    else:  # non-weight params at least in 16bit
        bw = 8
        print(f"INFO: Since {param} seems to be no weight tensor, it was quantized to {bw} bit instead of {bw_ap} bit.")

    prm = approx_data_in["parameters"][param]
    if approx_info["unsigned_integer_support"] and len(prm[prm < 0]) == 0:
        quantizedValues = quantizedValues.clip(0, 2 ** bw - 1)
    else:
        quantizedValues = quantizedValues.clip(-2 ** (bw - 1), 2 ** (bw - 1) - 1)
    return quantizedValues

def approx(approx_info, model_info, approx_data_in, enc_info):

    ##Qunatize tensor with uniform but without DQ
    approx_data_out = {k: copy.copy(v) for k, v in approx_data_in.items()} # create copies of dicts in approx_data
    encoder = deepCABAC.Encoder()
    model_access = NNRModelAccess(model_info)
    for block_or_param in model_access.blocks_and_params():
        for par_type, param, _ in block_or_param.param_generator(approx_data_in["compressed_parameter_types"]):
            if (par_type in approx_info["to_approximate"]) and (param not in approx_data_in["approx_method"]):
                # !!! There seems to be a pybind11 issue when using np.zeros_like for "values" that have been transposed.
                # !!! It seems that sometimes, encoder.quantLayer returns only zeros for quantizedValues. Needs further study.
                # !!! For now, using np.zeros instead of np.zeros_like seems to be a workaround.
                quantizedValues = np.zeros(approx_data_in["parameters"][param].shape, dtype=np.int32)
                encoder.initCtxModels( approx_info["cabac_unary_length_minus1"], 0 )
                
                qp_off = 0
                if approx_info['dq_flag'][param] == 1:
                    qp_off = common.compute_qp_offset_to_dq_equivalent( approx_data_out['qp_density'] )
                    print("INFO: Dependent quatization (DQ) can not be used with 'codebook'. In order to get similiar performance (to DQ) the QP is changed by {}!".format(-qp_off))

                enc_qp = approx_info['qp'][param] - qp_off

                qp = encoder.quantLayer(
                    approx_data_in["parameters"][param],
                    quantizedValues,
                    0, #approx_info['dq_flag'][param],
                    approx_data_out['qp_density'],
                    enc_qp,
                    approx_info["lambda_scale"],
                    approx_info["cabac_unary_length_minus1"],
                    approx_data_in["scan_order"].get(param, 0),
                    enc_info.get("general_profile_idc", 0) if enc_info else 0
                )

                if "integer_aligned_bitdepth" in approx_info and not check_array_all_zero_or_scalar(approx_data_out["parameters"][param]):
                    quantizedValues = clip_for_int_aligned(quantizedValues, approx_info, approx_data_in, model_info, param)
                if qp != enc_qp:
                    print("INFO: QP for {} has been clipped from {} to {} to avoid int32_t overflow!".format(param, approx_info['qp'][param],qp))
                    approx_data_out['qp'][param] = qp
                else:
                    approx_data_out['qp'][param] = enc_qp

                codebook, indexes = derive_sorted_codebook_from_tensor(quantizedValues)
                codebook, indexes, codebookOffset = get_codebook_offset( codebook, indexes,  approx_info["cabac_unary_length_minus1"])
                egk, _ = get_best_egk(codebook, codebookOffset)

                if approx_info["codebook_mode"] == 1:
                    approx_data_out["parameters"][param] = indexes
                    approx_data_out["codebooks"][param] = codebook
                    approx_data_out['approx_method'][param] = 'codebook'
                    approx_data_out['dq_flag'][param] = 0
                    approx_data_out["codebook_zero_offsets"][param] = codebookOffset
                    approx_data_out['codebooks_egk'][param] = egk
                
                elif approx_info["codebook_mode"] == 2:
                    if approx_info['dq_flag'][param] == 1:
                        quantizedValues = np.zeros(approx_data_in["parameters"][param].shape, dtype=np.int32)
                        encoder.initCtxModels( approx_info["cabac_unary_length_minus1"], 0 )
                        
                        enc_qp = approx_info['qp'][param]
                        ##else quantize again with DQ
                        qp = encoder.quantLayer(
                            approx_data_in["parameters"][param],
                            quantizedValues,
                            approx_info['dq_flag'][param],
                            approx_data_out['qp_density'],
                            enc_qp,
                            approx_info["lambda_scale"],
                            approx_info["cabac_unary_length_minus1"],
                            approx_data_in["scan_order"].get(param, 0),
                            enc_info.get("general_profile_idc", 0) if enc_info else 0,
                        )
                    if "integer_aligned_bitdepth" in approx_info and not check_array_all_zero_or_scalar(approx_data_out["parameters"][param]):
                        quantizedValues = clip_for_int_aligned(quantizedValues, approx_info, approx_data_in, model_info, param)

                    ##Compute Cost for encoding uniform quantized parameters
                    testEnc = deepCABAC.Encoder()
                    testEnc.initCtxModels( approx_info["cabac_unary_length_minus1"], enc_info.get("param_opt_flag", 0) if enc_info else 0 )
                    hdsp_opts = HDSP_OPTS_OFF()
                    testEnc.encodeLayer(quantizedValues, approx_info['dq_flag'][param], approx_data_in["scan_order"].get(param, 0),
                                        enc_info.get("general_profile_idc", 0) if enc_info else 0,
                                        enc_info.get('parent_node_id_present_flag', 0) if enc_info else 0,
                                        0, np.zeros(quantizedValues.shape[0], dtype=np.int32), *hdsp_opts, 0, 0
                                        )
                    bs_par = bytearray( testEnc.finish().tobytes() )

                    bytesUni = len(bs_par)
                    ##Compute cost for codebook quantized parameters + bytes for encoding the codebooks
                    testEnc = deepCABAC.Encoder()
                    testEnc.initCtxModels( approx_info["cabac_unary_length_minus1"], enc_info.get("param_opt_flag", 0) if enc_info else 0 )
                    testEnc.encodeLayer(indexes, 0, approx_data_in["scan_order"].get(param, 0),
                                        enc_info.get("general_profile_idc", 0) if enc_info else 0,
                                        enc_info.get('parent_node_id_present_flag', 0) if enc_info else 0,
                                        0, np.zeros(quantizedValues.shape[0], dtype=np.int32), *hdsp_opts, 0, 0
                                        )
                    bs_par_cb = bytearray( testEnc.finish().tobytes() )

                    bytesCb = len(bs_par_cb) + get_codebook_bytes(codebook, codebookOffset, egk)

                    ##select cheapest
                    if bytesCb < bytesUni:
                        approx_data_out["parameters"][param] = indexes
                        approx_data_out["codebooks"][param] = codebook
                        approx_data_out['approx_method'][param] = 'codebook'
                        approx_data_out['dq_flag'][param] = 0
                        approx_data_out["codebook_zero_offsets"][param] = codebookOffset
                        approx_data_out['codebooks_egk'][param] = egk
                    else:
                        print(f"INFO: Fallback to uniform quantization since results in smaller bitstream size ({bytesUni} bytes vs. {bytesCb} bytes)")
                        if approx_info['dq_flag'][param] == 1:
                            if qp != enc_qp:
                                print("INFO: QP for {} has been clipped from {} to {} to avoid int32_t overflow!".format(param, approx_info['qp'][param],qp))
                                approx_data_out['qp'][param] = qp
                            else:
                                approx_data_out['qp'][param] = enc_qp
                        approx_data_out['parameters'][param] = quantizedValues
                        approx_data_out['approx_method'][param] = 'uniform'
                        approx_data_out['dq_flag'][param] = approx_info['dq_flag'][param]

    approx_info_out = approx_info

    return approx_data_out, approx_info_out

# **********************************************************************************************************************
def rec(param, approx_data):
    assert approx_data['parameters'][param].dtype == np.int32
    cb = approx_data['codebooks'][param]
    stepsize = common.get_stepsize_from_qp(approx_data["qp"][param], approx_data["qp_density"])
    cb = cb * stepsize
    offset = approx_data['codebook_zero_offsets'][param]
    approx_data["parameters"][param] = cb[approx_data["parameters"][param] + offset]
    del approx_data["approx_method"][param]
    del approx_data['codebooks'][param]
    del approx_data['codebook_zero_offsets'][param]
    del approx_data['codebooks_egk'][param]
    del approx_data['qp'][param]
    



'''
Taken from llama.c code and lightly modified
Copyright (c) Andrej Karpathy
https://github.com/karpathy/llama2.c
under MIT License

The copyright in *this* software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or
other Intellectual Property Rights other than the copyrights concerning
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import math
import inspect
from dataclasses import dataclass
from typing import Any, Optional, Tuple
import torch
import torch.nn.functional as F
from torch import nn


@dataclass
class ModelArgs:
    # default hyperparameters for the Llama 7B model
    dim: int = 4096
    n_layers: int = 32
    n_heads: int = 32
    n_kv_heads: Optional[int] = None
    vocab_size: int = 32000
    hidden_dim: Optional[int] = None
    multiple_of: int = 256  # MLP hidden layer size will be multiple of
    norm_eps: float = 1e-5
    max_seq_len: int = 2048
    dropout: float = 0.0


class RMSNorm(torch.nn.Module):
    def __init__(self, dim: int, eps: float):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(dim))

    def _norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x):
        output = self._norm(x.float()).type_as(x)
        return output * self.weight


def precompute_freqs_cis(dim: int, end: int, theta: float = 10000.0):
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim))
    t = torch.arange(end, device=freqs.device)  # type: ignore
    freqs = torch.outer(t, freqs).float()  # type: ignore
    freqs_cos = torch.cos(freqs)  # real part
    freqs_sin = torch.sin(freqs)  # imaginary part
    return freqs_cos, freqs_sin

def reshape_for_broadcast(freqs_cis: torch.Tensor, x: torch.Tensor):
    ndim = x.ndim
    assert 0 <= 1 < ndim
    assert freqs_cis.shape == (x.shape[1], x.shape[-1])
    shape = [d if i == 1 or i == ndim - 1 else 1 for i, d in enumerate(x.shape)]
    return freqs_cis.view(shape)

def apply_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_cos: torch.Tensor,
    freqs_sin: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:

    # reshape xq and xk to match the complex representation
    xq_r, xq_i = xq.float().reshape(xq.shape[:-1] + (-1, 2)).unbind(-1)
    xk_r, xk_i = xk.float().reshape(xk.shape[:-1] + (-1, 2)).unbind(-1)

    # reshape freqs_cos and freqs_sin for broadcasting
    freqs_cos = reshape_for_broadcast(freqs_cos, xq_r)
    freqs_sin = reshape_for_broadcast(freqs_sin, xq_r)

    # apply rotation using real numbers
    xq_out_r = xq_r * freqs_cos - xq_i * freqs_sin
    xq_out_i = xq_r * freqs_sin + xq_i * freqs_cos
    xk_out_r = xk_r * freqs_cos - xk_i * freqs_sin
    xk_out_i = xk_r * freqs_sin + xk_i * freqs_cos

    # flatten last two dimensions
    xq_out = torch.stack([xq_out_r, xq_out_i], dim=-1).flatten(3)
    xk_out = torch.stack([xk_out_r, xk_out_i], dim=-1).flatten(3)

    return xq_out.type_as(xq), xk_out.type_as(xk)

def repeat_kv(x: torch.Tensor, n_rep: int) -> torch.Tensor:
    """torch.repeat_interleave(x, dim=2, repeats=n_rep)"""
    bs, slen, n_kv_heads, head_dim = x.shape
    if n_rep == 1:
        return x
    return (
        x[:, :, :, None, :]
        .expand(bs, slen, n_kv_heads, n_rep, head_dim)
        .reshape(bs, slen, n_kv_heads * n_rep, head_dim)
    )

class Attention(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.n_kv_heads = args.n_heads if args.n_kv_heads is None else args.n_kv_heads
        assert args.n_heads % self.n_kv_heads == 0
        model_parallel_size = 1
        self.n_local_heads = args.n_heads // model_parallel_size
        self.n_local_kv_heads = self.n_kv_heads // model_parallel_size
        self.n_rep = self.n_local_heads // self.n_local_kv_heads
        self.head_dim = args.dim // args.n_heads
        self.wq = nn.Linear(args.dim, args.n_heads * self.head_dim, bias=False)
        self.wk = nn.Linear(args.dim, self.n_kv_heads * self.head_dim, bias=False)
        self.wv = nn.Linear(args.dim, self.n_kv_heads * self.head_dim, bias=False)
        self.wo = nn.Linear(args.n_heads * self.head_dim, args.dim, bias=False)
        self.attn_dropout = nn.Dropout(args.dropout)
        self.resid_dropout = nn.Dropout(args.dropout)
        self.dropout = args.dropout

        # use flash attention or a manual implementation?
        self.flash = hasattr(torch.nn.functional, 'scaled_dot_product_attention')
        if not self.flash:
            print("WARNING: using slow attention. Flash Attention requires PyTorch >= 2.0")
            mask = torch.full((1, 1, args.max_seq_len, args.max_seq_len), float("-inf"))
            mask = torch.triu(mask, diagonal=1)
            self.register_buffer("mask", mask)

    def forward(
        self,
        x: torch.Tensor,
        freqs_cos: torch.Tensor,
        freqs_sin: torch.Tensor,
    ):
        bsz, seqlen, _ = x.shape

        # QKV
        xq, xk, xv = self.wq(x), self.wk(x), self.wv(x)
        xq = xq.view(bsz, seqlen, self.n_local_heads, self.head_dim)
        xk = xk.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)
        xv = xv.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)

        # RoPE relative positional embeddings
        xq, xk = apply_rotary_emb(xq, xk, freqs_cos, freqs_sin)

        # grouped multiquery attention: expand out keys and values
        xk = repeat_kv(xk, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)
        xv = repeat_kv(xv, self.n_rep)  # (bs, seqlen, n_local_heads, head_dim)

        # make heads into a batch dimension
        xq = xq.transpose(1, 2)  # (bs, n_local_heads, seqlen, head_dim)
        xk = xk.transpose(1, 2)
        xv = xv.transpose(1, 2)

        # flash implementation
        if self.flash:
            output = torch.nn.functional.scaled_dot_product_attention(xq, xk, xv, attn_mask=None, dropout_p=self.dropout if self.training else 0.0, is_causal=True)
        else:
            # manual implementation
            scores = torch.matmul(xq, xk.transpose(2, 3)) / math.sqrt(self.head_dim)
            assert hasattr(self, 'mask')
            scores = scores + self.mask[:, :, :seqlen, :seqlen]   # (bs, n_local_heads, seqlen, cache_len + seqlen)
            scores = F.softmax(scores.float(), dim=-1).type_as(xq)
            scores = self.attn_dropout(scores)
            output = torch.matmul(scores, xv)  # (bs, n_local_heads, seqlen, head_dim)

        # restore time as batch dimension and concat heads
        output = output.transpose(1, 2).contiguous().view(bsz, seqlen, -1)

        # final projection into the residual stream
        output = self.wo(output)
        output = self.resid_dropout(output)
        return output


class FeedForward(nn.Module):
    def __init__(self, dim: int, hidden_dim: int, multiple_of: int, dropout: float):
        super().__init__()
        if hidden_dim is None:
            hidden_dim = 4 * dim
            hidden_dim = int(2 * hidden_dim / 3)
            hidden_dim = multiple_of * ((hidden_dim + multiple_of - 1) // multiple_of)
        self.w1 = nn.Linear(dim, hidden_dim, bias=False)
        self.w2 = nn.Linear(hidden_dim, dim, bias=False)
        self.w3 = nn.Linear(dim, hidden_dim, bias=False)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.dropout(self.w2(F.silu(self.w1(x)) * self.w3(x)))

class BinClassifier(nn.Module):
    def __init__(self, dim: int, hidden_dim: int):
        super().__init__()
        self.w1 = nn.Linear(dim, hidden_dim, bias=True)
        self.w2 = nn.Linear(hidden_dim, hidden_dim//2, bias=True)
        self.w3 = nn.Linear(hidden_dim//2, 2, bias=True)

    def forward(self, x):
        return self.w3(F.silu(self.w2(F.silu(self.w1(x)))))


class TransformerBlock(nn.Module):
    def __init__(self, layer_id: int, args: ModelArgs):
        super().__init__()
        self.n_heads = args.n_heads
        self.dim = args.dim
        self.head_dim = args.dim // args.n_heads
        self.attention = Attention(args)
        self.feed_forward = FeedForward(
            dim=args.dim,
            hidden_dim=args.hidden_dim,
            multiple_of=args.multiple_of,
            dropout=args.dropout,
        )
        self.layer_id = layer_id
        self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)

    def forward(self, x, freqs_cos, freqs_sin):
        h = x + self.attention.forward(self.attention_norm(x), freqs_cos, freqs_sin)
        out = h + self.feed_forward.forward(self.ffn_norm(h))
        return out


class Transformer(nn.Module):
    last_loss: Optional[torch.Tensor]

    def __init__(self, params: ModelArgs):
        super().__init__()
        self.params = params
        self.vocab_size = params.vocab_size
        self.n_layers = params.n_layers

        self.tok_embeddings = nn.Embedding(params.vocab_size, params.dim)
        self.dropout = nn.Dropout(params.dropout)
        self.layers = torch.nn.ModuleList()
        for layer_id in range(params.n_layers):
            self.layers.append(TransformerBlock(layer_id, params))
        self.norm = RMSNorm(params.dim, eps=params.norm_eps)
        self.output = nn.Linear(params.dim, params.vocab_size, bias=False)

        # share the unembedding parameters with the embedding parameters
        self.tok_embeddings.weight = self.output.weight # https://paperswithcode.com/method/weight-tying

        # some useful precompute for the RoPE relative positional embeddings
        freqs_cos, freqs_sin = precompute_freqs_cis(self.params.dim // self.params.n_heads, self.params.max_seq_len)
        self.register_buffer("freqs_cos", freqs_cos, persistent=False)
        self.register_buffer("freqs_sin", freqs_sin, persistent=False)

        # init all weights
        self.apply(self._init_weights)
        # apply special scaled init to the residual projections, per GPT-2 paper
        for pn, p in self.named_parameters():
            if pn.endswith('w3.weight') or pn.endswith('wo.weight'):
                torch.nn.init.normal_(p, mean=0.0, std=0.02/math.sqrt(2 * params.n_layers))

        # Initialize attribute for the loss of the last forward call. This will be set if the forward is called with a targets tensor.
        self.last_loss = None

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias)
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02)

    def forward(self, tokens: torch.Tensor,
                targets: Optional[torch.Tensor] = None) -> Optional[torch.Tensor]:

        _bsz, seqlen = tokens.shape
        h = self.tok_embeddings(tokens)
        h = self.dropout(h)
        freqs_cos = self.freqs_cos[:seqlen]
        freqs_sin = self.freqs_sin[:seqlen]

        for layer in self.layers:
            h = layer(h, freqs_cos, freqs_sin)
        h = self.norm(h)

        if targets is not None:
            # if we are given some desired targets also calculate the loss
            logits = self.output(h)
            if targets.dtype != torch.long:
                targets = targets.type(torch.long)
            self.last_loss = F.cross_entropy(logits.view(-1, logits.size(-1)), targets.view(-1), ignore_index=-1)
        else:
            # inference-time mini-optimization: only forward the output on the very last position
            logits = self.output(h[:, [-1], :]) # note: using list [-1] to preserve the time dim
            self.last_loss = None

        return logits

    def configure_optimizers(self, weight_decay, learning_rate, betas, device_type):
        # start with all of the candidate parameters
        param_dict = {pn: p for pn, p in self.named_parameters()}
        # filter out those that do not require grad
        param_dict = {pn: p for pn, p in param_dict.items() if p.requires_grad}
        # create optim groups. Any parameters that is 2D will be weight decayed, otherwise no.
        # i.e. all weight tensors in matmuls + embeddings decay, all biases and layernorms don't.
        decay_params = [p for n, p in param_dict.items() if p.dim() >= 2]
        nodecay_params = [p for n, p in param_dict.items() if p.dim() < 2]
        optim_groups = [
            {'params': decay_params, 'weight_decay': weight_decay},
            {'params': nodecay_params, 'weight_decay': 0.0}
        ]
        num_decay_params = sum(p.numel() for p in decay_params)
        num_nodecay_params = sum(p.numel() for p in nodecay_params)
        print(f"num decayed parameter tensors: {len(decay_params)}, with {num_decay_params:,} parameters ({num_decay_params*4/1e6:.4f} MB)")

        print(f"num non-decayed parameter tensors: {len(nodecay_params)}, with {num_nodecay_params:,} parameters ({num_nodecay_params*4/1e6:.4f} MB)")
        # Create AdamW optimizer and use the fused version if it is available
        fused_available = 'fused' in inspect.signature(torch.optim.AdamW).parameters
        use_fused = fused_available and device_type == 'cuda'
        extra_args = dict(fused=True) if use_fused else dict()
        optimizer = torch.optim.AdamW(optim_groups, lr=learning_rate, betas=betas, **extra_args)
        print(f"using fused AdamW: {use_fused}")

        return optimizer



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

import copy, logging
LOGGER = logging.getLogger()
from nncodec import nnc_core
import h5py
import os
# import tensorflow as tf

import copy, logging
import numpy as np
from nncodec.framework.use_case_init import use_cases
from nncodec.framework.applications.utils import evaluation
from collections import OrderedDict

def is_tef_model(model_object):
    try:
        import tensorflow as tf
        return isinstance(model_object, tf.Module)
    except ImportError:
        return False

def save_to_tensorflow_file( model_data, path ):
        h5_model = h5py.File(path, 'w')

        grp_names = []
        for module_name in model_data:
            splits = module_name.split('/')
            grp_name = module_name.split('/')[0]
            if splits[0] == splits[2]:
                grp_name = (splits[0] + '/' + splits[1])
            if grp_name not in grp_names:
                grp_names.append(grp_name)
            h5_model.create_dataset(module_name, data=model_data[module_name])
        h5_model.attrs['layer_names'] = grp_names

        for grp in h5_model:
            weight_attr = []
            if isinstance(h5_model[grp], h5py.Group) and grp in grp_names:
                weight_attr = [k[len(grp)+1:] for k, v in model_data.items()
                               if k.startswith(grp+'/')]
                h5_model[grp].attrs['weight_names'] = weight_attr
            elif isinstance(h5_model[grp], h5py.Group):
                for subgrp in h5_model[grp]:
                    if isinstance(h5_model[grp], h5py.Group) and (grp + '/' + subgrp) in grp_names:
                        weight_attr = [k[len(grp) + len(subgrp) + 2:] for k, v in model_data.items()
                                       if k.startswith(grp + '/' + subgrp + '/')]
                    h5_model[grp + '/' + subgrp].attrs['weight_names'] = weight_attr

def __initialize_data_functions(
                 handler=None,
                 dataset_path=None,
                 batch_size=None,
                 num_workers=None,
                 model_name='unspecified'
                ):

    if model_name==None:
        model_name='unspecified'

    if dataset_path:
        test_set, test_loader = handler.init_test_tef(
            dataset_path,
            batch_size,
            num_workers,
            model_name
        )

        val_set, val_loader = handler.init_validation_tef(
            dataset_path,
            batch_size,
            num_workers,
            model_name
        )

    else:
        test_set, test_loader = None, None
        val_set, val_loader   = None, None
    
    return test_set, test_loader, val_set, val_loader

def create_NNC_model_instance_from_file(
                 model_path,
                 dataset_path=None,
                 lr=1e-5,
                 batch_size=64,
                 num_workers=1,
                 model_struct=None,
                 model_name=None
                ):

    TEFModel = TensorFlowModel()
    model_parameters, loaded_model_struct = TEFModel.load_model(model_path)
    
    if model_struct == None and loaded_model_struct != None:
        model_struct = loaded_model_struct

    if dataset_path and model_struct:
        if model_name == None and hasattr(model_struct, 'name'):
            model_name=model_struct.name
        TEFModelExecuter = create_imagenet_model_executer(model_struct=model_struct,
                                                          dataset_path=dataset_path,
                                                          lr=lr,
                                                          batch_size=batch_size,
                                                          num_workers=num_workers,
                                                          model_name=model_name)

    else:
        TEFModelExecuter = None

    return TEFModel, TEFModelExecuter, model_parameters


def create_NNC_model_instance_from_object(
                 model_object,
                 dataset_path=None,
                 lr=1e-5,
                 batch_size=64,
                 num_workers=1,
                 model_struct=None,
                 model_name=None
                ):

    TEFModel = TensorFlowModel()
    model_parameters, loaded_model_struct = TEFModel.init_model_from_model_object(model_object)
    
    if model_struct == None and loaded_model_struct != None:
        model_struct = loaded_model_struct

    if dataset_path and model_struct:
        if model_name == None and hasattr(model_struct, 'name'):
            model_name=model_struct.name
        TEFModelExecuter = create_imagenet_model_executer(model_struct=model_struct,
                                                          dataset_path=dataset_path,
                                                          lr=lr,
                                                          batch_size=batch_size,
                                                          num_workers=num_workers,
                                                          model_name=model_name)

    else:
        TEFModelExecuter = None

    return TEFModel, TEFModelExecuter, model_parameters


def create_imagenet_model_executer( 
                            model_struct,
                            dataset_path,
                            lr=1e-5,
                            batch_size=64,
                            num_workers=1,
                            model_name=None,
                            ):
    
    assert model_struct != None, "model_struct must be specified in order to create a model_executer!"
    assert dataset_path != None, "dataset_path must be specified in order to create a model_executer!"
    
    handler = use_cases['NNR_TEF']

    test_set, test_loader, val_set, val_loader = __initialize_data_functions( handler=handler,
                                                                              dataset_path=dataset_path,
                                                                              batch_size=batch_size,
                                                                              num_workers=num_workers,
                                                                              model_name=model_name)

    assert (test_set!=None and test_loader!= None) or ( val_set!= None and val_loader!= None ), "Any of the pairs test_set/test_loader or val_set/val_loader must be specified in order to use data driven optimizations methods!"
    TEFModelExecuter = ImageNetTensorFlowModelExecuter( handler,
                                                        train_loader=None,
                                                        test_loader=test_loader,
                                                        test_set=test_set,
                                                        val_loader=val_loader,
                                                        val_set=val_set,
                                                        model_struct=model_struct)

    return TEFModelExecuter


def get_model_file_with_parameters( parameters, model_struct ):
    new_model_struct = copy.deepcopy(model_struct)
    
    save_to_tensorflow_file(parameters, './intermed_h5_model.h5')
    new_model_struct.load_weights('./intermed_h5_model.h5')
    os.remove('./intermed_h5_model.h5')

    return new_model_struct


class TensorFlowModel(nnc_core.nnr_model.NNRModel):

    def __init__(self, model_dict=None):
        
        if model_dict and isinstance(model_dict, dict):
            self.init_model_from_dict(model_dict)
        else:
            self.__model_info = None

        self.model = None

    def load_model( self, 
                    model_path
                  ):

        try:
            model_file = tf.keras.models.load_model(model_path)
        except:
            model_file = h5py.File(model_path, 'r')
        
        try:
            if isinstance(model_file, dict):
                model_parameter_dict = model_file
            elif isinstance(model_file, tf.Module):
                return self.init_model_from_model_object(model_file)
            else:
                if 'layer_names' in model_file.attrs:
                    module_names = [n for n in model_file.attrs['layer_names']]

                layer_names = []
                for mod_name in module_names:
                    layer = model_file[mod_name]
                    if 'weight_names' in layer.attrs:
                        weight_names = [mod_name+'/'+n for n in layer.attrs['weight_names']]
                        if weight_names:
                            layer_names += weight_names

                model_parameter_dict = {}
                for name in layer_names:
                    model_parameter_dict[name] = model_file[name]
        except:
            raise SystemExit("Can't read model: {}".format(model_path))

        return self.init_model_from_dict( model_parameter_dict ), None


    def init_model_from_model_object(   self,
                                        model_object,
                                    ):
        self.model = model_object

        # h5_model_path = './temp.h5'
        # model_object.save_weights(h5_model_path)
        # model = h5py.File(h5_model_path, 'r')
        # os.remove(h5_model_path)
        #
        # if 'layer_names' in model.attrs:
        #     module_names = [n for n in model.attrs['layer_names']]
        #
        # layer_names = []
        # for mod_name in module_names:
        #     layer = model[mod_name]
        #     if 'weight_names' in layer.attrs:
        #         weight_names = [mod_name+'/'+n for n in layer.attrs['weight_names']]
        #         if weight_names:
        #             layer_names += weight_names
        #
        # model_parameter_dict = {}
        # for name in layer_names:
        #     model_parameter_dict[name] = model[name]

        weights = model_object.get_weights()
        layer_names = []

        for layer in model_object.layers:
            mod_name = layer.name
            if layer.weights != []:
                for weight in layer.weights:
                    layer_names.append(mod_name + "/" + weight.name)

        model_parameter_dict = {}
        for i, name in enumerate(layer_names):
            model_parameter_dict[name] = weights[i]

        return self.init_model_from_dict( model_parameter_dict ), model_object


    @property
    def model_info(self):
        return self.__model_info


    def init_model_from_dict(self, tf_dict):


        model_data = {'parameters': {}, 'reduction_method': 'baseline'}
        model_info = {'parameter_type': {}, 'parameter_dimensions': {}, 'parameter_index': {}, 'block_identifier': {}, 'original_size': {}, 'topology_storage_format' : None, 'topology_compression_format' : None}

        type_list_int = ['int8', 'int16', 'int32', 'uint8', 'uint16', 'uint32']
        type_list_1_bytes = ['int8', 'uint8']
        type_list_2_bytes = ['int16', 'uint16', 'float16']
        original_size = 0

        for i, module_name in enumerate(tf_dict):
            model_data['parameters'][module_name] = tf_dict[module_name][()]
            if  model_data['parameters'][module_name].dtype in type_list_1_bytes:
                original_size += model_data['parameters'][module_name].size
            elif  model_data['parameters'][module_name].dtype in type_list_2_bytes:
                original_size +=  model_data['parameters'][module_name].size*2
            else:
                original_size +=  model_data['parameters'][module_name].size*4
            model_data['parameters'][module_name] = np.int32(model_data['parameters'][module_name]) if model_data['parameters'][module_name].dtype in type_list_int else model_data['parameters'][module_name]
            mdl_shape = model_data['parameters'][module_name].shape
            model_info['parameter_dimensions'][module_name] = mdl_shape
            if len(mdl_shape) == 0: #scalar
                model_data['parameters'][module_name] = np.array([np.float32(tf_dict[module_name][()])])
                model_info['parameter_dimensions'][module_name] = np.array([0]).shape
            model_info['parameter_index'][module_name] = i

            dims = len(mdl_shape)

            quantize = True
            quantize_onedim = True

            if dims > 1 and 'kernel' in module_name:
                model_info['parameter_type'][module_name] = 'weight' if quantize else 'unspecified'
            elif dims > 1:
                model_info['parameter_type'][module_name] = 'weight' if quantize else 'unspecified'
            elif dims == 1:
                if 'bias' in module_name:
                    model_info['parameter_type'][module_name] = 'bias' if quantize_onedim else 'unspecified'
                elif 'beta' in module_name:
                    model_info['parameter_type'][module_name] = 'bn.beta' if quantize_onedim else 'unspecified'
                elif 'moving_mean' in module_name:
                    model_info['parameter_type'][module_name] = 'bn.mean' if quantize_onedim else 'unspecified'
                elif 'moving_variance' in module_name:
                    model_info['parameter_type'][module_name] = 'bn.var' if quantize_onedim else 'unspecified'
                elif 'gamma' in module_name:
                    model_info['parameter_type'][module_name] = 'bn.gamma' if quantize_onedim else 'unspecified'
                elif 'weight' in module_name:
                    model_info['parameter_type'][module_name] = 'weight' if quantize_onedim else 'unspecified'
                else:
                    model_info['parameter_type'][module_name] = 'unspecified'
            else:
                model_info['parameter_type'][module_name] = 'unspecified'

        model_info['topology_storage_format'] = nnc_core.nnr_model.TopologyStorageFormat.NNR_TPL_TEF
        model_info['topology_compression_format'] = nnc_core.nnr_model.TopologyCompressionFormat.NNR_PT_RAW

        model_info["original_size"] = original_size
        
        self.__model_info = model_info

        return model_data["parameters"]

    def save_state_dict(self, path, model_data):
        h5_model = h5py.File(path, 'w')

        grp_names = []
        for module_name in model_data:
            splits = module_name.split('/')
            grp_name = module_name.split('/')[0].encode('utf8')
            if splits[0] == splits[2]:
                grp_name = (splits[0] + '/' + splits[1]).encode('utf8')
            if grp_name not in grp_names:
                grp_names.append(grp_name)
            if model_data[module_name].size != 1:
                h5_model.create_dataset(module_name, data=model_data[module_name])
            else: #scalar
                h5_model.create_dataset(module_name, data=np.int64(model_data[module_name][0]))
        h5_model.attrs['layer_names'] = grp_names

        for grp in h5_model:
            weight_attr = []
            if isinstance(h5_model[grp], h5py.Group) and grp.encode('utf8') in grp_names:
                weight_attr = [k[len(grp)+1:].encode('utf8') for k, v in model_data.items()
                               if k.startswith(grp+'/')]
                h5_model[grp].attrs['weight_names'] = weight_attr
            elif isinstance(h5_model[grp], h5py.Group):
                for subgrp in h5_model[grp]:
                    if isinstance(h5_model[grp], h5py.Group) and (grp + '/' + subgrp).encode('utf8') in grp_names:
                        weight_attr = [k[len(grp) + len(subgrp) + 2:].encode('utf8') for k, v in model_data.items()
                                       if k.startswith(grp + '/' + subgrp + '/')]
                    h5_model[grp + '/' + subgrp].attrs['weight_names'] = weight_attr

    
    def get_model_struct(self):
        return self.model


    def guess_block_id_and_param_type(self, model_parameters):
        try:
            block_id_and_param_type = {"block_identifier" : {}, "parameter_type" : {}}
            block_dict = dict()
            blkNum = -1
            for param in model_parameters.keys():
                splitted_param = param.split("/")
                param_end = splitted_param[-1]
                base_block_id  = "/".join(splitted_param[0:-1])
                base_block_id  = "/".join(splitted_param[0:-1])+":" if len(splitted_param[0:-1]) != 0 else "genericBlk:"
                dims = len(model_parameters[param].shape)
                paramShape = model_parameters[param].shape
            

                if dims > 1 and ('kernel' in param_end or 'weight' in param_end):
                    paramType = 'weight'
                    blockId = base_block_id
                elif dims > 1:
                    paramType = 'weight'
                    blockId = base_block_id
                elif dims == 1:
                    if 'bias' in param_end or 'beta' in param_end: ##could also be bn.beta
                        paramType = 'bias'
                        blockId = base_block_id
                    elif 'running_mean' in param_end or 'moving_mean' in param_end:
                        paramType = 'bn.mean'
                        blockId = base_block_id
                    elif 'running_var' in param_end or 'moving_variance' in param_end:                        
                        paramType = 'bn.var'
                        blockId = base_block_id
                    elif 'weight_scaling' in param_end:
                        paramType = 'weight.ls'
                        blockId = base_block_id
                    elif 'gamma' in param_end:
                        paramType = 'bn.gamma'
                        blockId = base_block_id
                    elif 'weight' in param_end:
                        paramType = 'weight'
                        blockId = base_block_id
                    else:
                        paramType = 'unspecified'
                        blockId = None
                else:
                    paramType = 'unspecified'
                    blockId = None
                
                
                if blockId:
                    block_id = base_block_id + str(blkNum)
                    if block_id in block_dict.keys():
                        if any([a[1] == paramType for a in block_dict[block_id]]):
                            blkNum += 1
                        block_id = base_block_id + str(blkNum)
                        blockId = base_block_id + str(blkNum)
                    else:
                        blkNum += 1
                        block_id = base_block_id + str(blkNum)
                        blockId = base_block_id + str(blkNum)
                        
                    if block_id not in block_dict.keys():
                        block_dict[block_id] = []
                        
                    block_dict[block_id].append( [param, paramType, blockId, dims, paramShape] )
                else:
                    block_id_and_param_type["parameter_type"][param] = paramType
                    block_id_and_param_type["block_identifier"][param] = blockId
                
                
            weight_block_list = []
            bn_block_list     = []

            for block_list in block_dict.values():
                if any(["bn." in a[1] for a in block_list]):
                    for i, val in enumerate(block_list):
                        par, parT, blkId, dims, _ = val
                        if parT == 'weight' and dims == 1:
                            block_list[i][1] = "bn.gamma"
                        if parT == 'bias':
                            block_list[i][1] = "bn.beta"
                    bn_block_list.append( block_list )
                else:
                    weight_block_list.append(block_list)
                    
            weight_shape = None
            weight_blkId = None
            for weight_block in weight_block_list:
                weight_shape = None
                weight_blkId = None
                for par, parT, blkId, dims, paramSh in weight_block:
                    block_id_and_param_type["parameter_type"][par] = parT
                    block_id_and_param_type["block_identifier"][par] = blkId
                    if parT == 'weight':
                        weight_shape = paramSh
                        weight_blkId = blkId
            
                if len(bn_block_list) != 0 and any([dim == bn_block_list[0][0][4][0] for dim in weight_shape]):
                    bn_block = bn_block_list.pop(0)
                    for par, parT, _, _, _ in bn_block:
                        block_id_and_param_type["parameter_type"][par] = parT
                        block_id_and_param_type["block_identifier"][par] = weight_blkId
            
            assert len(bn_block_list) == 0, "Unhandled BN parameters!"
            
        except:
            print("INFO: Guessing of block_id_and_parameter_type failed! block_id_and_parameter_type has been set to 'None'!")
            block_id_and_param_type = None

        return block_id_and_param_type



class ImageNetTensorFlowModelExecuter( nnc_core.nnr_model.ModelExecute ):

    def __init__(self,
                 handler,
                 train_loader,
                 test_loader,
                 test_set,
                 val_loader,
                 val_set,
                 model_struct):
        
        self.handle = handler
        if test_set:
            self.test_set = test_set
            self.test_loader = test_loader
        if val_set:
            self.val_set = val_set
            self.val_loader = val_loader
        if train_loader:
            self.train_loader = train_loader

        if model_struct:
            self.original_model = model_struct
            self.model = model_struct
        else:
            self.original_model = None
            self.model = None

    
    def save_state_dict(self, path, model_data):
        h5_model = h5py.File(path, 'w')

        grp_names = []
        for module_name in model_data:
            splits = module_name.split('/')
            grp_name = module_name.split('/')[0].encode('utf8')
            if splits[0] == splits[2]:
                grp_name = (splits[0] + '/' + splits[1]).encode('utf8')
            if grp_name not in grp_names:
                grp_names.append(grp_name)
            if model_data[module_name].size != 1:
                h5_model.create_dataset(module_name, data=model_data[module_name])
            else: #scalar
                h5_model.create_dataset(module_name, data=np.int64(model_data[module_name][0]))
        h5_model.attrs['layer_names'] = grp_names

        for grp in h5_model:
            weight_attr = []
            if isinstance(h5_model[grp], h5py.Group) and grp.encode('utf8') in grp_names:
                weight_attr = [k[len(grp)+1:].encode('utf8') for k, v in model_data.items()
                               if k.startswith(grp+'/')]
                h5_model[grp].attrs['weight_names'] = weight_attr
            elif isinstance(h5_model[grp], h5py.Group):
                for subgrp in h5_model[grp]:
                    if isinstance(h5_model[grp], h5py.Group) and (grp + '/' + subgrp).encode('utf8') in grp_names:
                        weight_attr = [k[len(grp) + len(subgrp) + 2:].encode('utf8') for k, v in model_data.items()
                                       if k.startswith(grp + '/' + subgrp + '/')]
                    h5_model[grp + '/' + subgrp].attrs['weight_names'] = weight_attr

    def test_model(
            self,
            parameters,
            verbose=False
    ):
            tf.config.optimizer.set_jit(True)

            Model = self.model
            self.save_state_dict('./intermed_h5_model.h5', parameters)
            Model.load_weights('./intermed_h5_model.h5')
            os.remove('./intermed_h5_model.h5')

            acc = self.handle.evaluate( 
            Model,
            self.test_loader,
            self.test_set,
            num_workers=8,
            verbose=verbose
        )
            del Model   
            return acc

    def eval_model(
            self,
            parameters,
            verbose=False,
    ):

            tf.config.optimizer.set_jit(True)

            Model = self.model
            self.save_state_dict('./intermed_h5_model.h5', parameters)
            Model.load_weights('./intermed_h5_model.h5')
            os.remove('./intermed_h5_model.h5')

            acc = self.handle.evaluate( 
            Model,
            self.val_loader,
            self.val_set,
            num_workers=8,
            verbose=verbose
        )
            del Model   
            return acc

    def has_eval(self):
        return True
    
    def has_test(self):
        return True
    
    def has_tune_ft(self):
        return False
    
    def has_tune_lsa(self):
        return False

        


'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import logging
import torch
import time
import random
import numpy as np
from tqdm import tqdm
from contextlib import nullcontext
from nncodec.nnc_core import nnr_model
import math

LOGGER = logging.getLogger(__name__)

def freeze_batch_norm_layers(model):
    for name, mod in model.named_modules():
        if isinstance(mod, torch.nn.BatchNorm2d):
            mod.eval()

def train_classification_model(model, optimizer=None, criterion=None, trainloader=None, device=None, verbose=True,
                               max_batches=None, freeze_batch_norm=False, return_model=False, args=None, round=0):
    """
    Parameters
    ----------
    device: torch.device
        Choose between cuda or cpu.
    model: torch.nn.Module
        A pytorch network model.
    optimizer: torch.optim.Optimizer
        A pytorch optimizer like Adam.
    criterion: torch.nn.Loss
        A pytorch criterion that defines the loss.
    trainloader: torch.utils.data.DataLoader
        Loader of train data.
    max_batches: int
        How many batches the model should train for.
    verbose: bool
        If True, print text - verbose mode.
    freeze_batch_norm: bool
        If True set batch norm layers to eval. Default: False

    Returns
    -------
    success: bool
        Returns False is nans encountered in the loss else True.
    """

    if isinstance(model, nnr_model.ModelExecute):
        optimizer = model.optimizer
        criterion = model.handle.criterion
        trainloader = model.test_loader
        device = model.device
        model = model.model

    if criterion is None:
        criterion = torch.nn.CrossEntropyLoss()

    model.to(device)
    model.train()
    if freeze_batch_norm:
        freeze_batch_norm_layers(model)

    train_loss = []
    correct = 0
    total = 0

    if args is not None and args.max_batches == None:
        max_batches = len(trainloader)
    elif args is not None:
        max_batches = args.max_batches


    total_iterations = max_batches or len(trainloader)
    iterator = tqdm(enumerate(trainloader), total=total_iterations, position=0, leave=True, desc='train_classification') \
        if verbose else enumerate(trainloader)

    DeepLab_condition = model.__class__.__name__ == "DeepLabV3" if not (isinstance(model, torch.nn.DataParallel) or isinstance(model, torch.nn.parallel.DistributedDataParallel)) \
                            else model.module.__class__.__name__ == "DeepLabV3"

    if DeepLab_condition:
        from torchmetrics import JaccardIndex
        num_of_classes = model.classifier[len(model.classifier) - 1].weight.shape[0]
        jaccard = JaccardIndex(task="multiclass", num_classes=num_of_classes, average="macro").to(device)

    for batch_idx, (inputs, targets) in iterator:

        inputs, targets = inputs.to(device), targets.to(device)
        optimizer.zero_grad()
        outputs = model(inputs)

        if DeepLab_condition:
            outputs = outputs['out']
            targets = targets * (targets != 1) * 255 # for Pascal VOC (different for, e.g., RailSem19)
            targets = targets.squeeze(1).long() # for Pascal VOC (different for, e.g., RailSem19 (use "decode_mask_RS()")

        loss = criterion(outputs, targets)

        if torch.isnan(loss):
            LOGGER.warning('--> Loss is Nan.')
            break

        loss.backward()
        optimizer.step()
        train_loss.append(loss.item())

        _, predicted = outputs.max(1)

        total += targets.size(0) if not DeepLab_condition else targets.numel()
        correct += predicted.eq(targets).sum().item()

        if DeepLab_condition:
            jaccard.update(predicted, targets)

        if batch_idx == max_batches:
            break

    acc = correct * 100.0 / total
    mean_train_loss = np.mean(train_loss)

    if return_model:
        if DeepLab_condition:
            final_mIoU = jaccard.compute() * 100
            return {'acc': acc, 'final_mIoU': final_mIoU.item()}, model
        else:
            return {'acc': acc, 'mean_train_loss': mean_train_loss}, model
    else:
        return {'acc': acc, 'mean_train_loss': mean_train_loss}


def train_language_model(model, optimizer=None, criterion=None, trainloader=None, device=None, verbose=False, args=None,
                         round=0):

    torch.manual_seed(808 + round)
    np.random.seed(303 + round)
    random.seed(909 + round)

    try:
        len_trainloader = len(trainloader)
    except TypeError:
        len_trainloader = trainloader.dataset.num_samples

    if args is not None and args.max_batches is None:
        max_batches = len_trainloader
    else:
        max_batches = args.max_batches if args is not None else int(1e6)

    model.to(device)
    model.train()

    class Iterator:
        @staticmethod
        def iter_batches(dataloader, device):
            for x, y in dataloader:
                x = x.to(device)
                y = y.to(device)
                yield x, y

    dtype = "float16"

    train_batches = []
    batch_iter = Iterator.iter_batches(trainloader, device)
    for idx, (x, y) in enumerate(batch_iter):
        if idx >= max_batches:
            break
        train_batches.append((x, y))

    print(f"[INFO] Loaded {len(train_batches)} batches for round {round}")

    iter_num = round * len(train_batches)
    log_interval = 100
    decay_lr = True
    warmup_iters = 1000
    lr_decay_iters = args.epochs * max_batches if args is not None else 10000
    min_lr = 0.0
    grad_clip = 1.0

    # Learning rate scheduler
    def get_lr(it):
        if it < warmup_iters:
            return args.lr * it / warmup_iters
        if it > lr_decay_iters:
            return min_lr
        decay_ratio = (it - warmup_iters) / (lr_decay_iters - warmup_iters)
        coeff = 0.5 * (1.0 + math.cos(math.pi * decay_ratio))
        return min_lr + coeff * (args.lr - min_lr)

    # initialize a GradScaler. If enabled=False scaler is a no-op
    scaler = torch.amp.GradScaler(enabled=(dtype == "float16" and torch.cuda.is_available()))

    train_loss = []
    train_acc = []
    t0 = time.time()
    ctx = nullcontext()

    for X, Y in train_batches:
        with ctx:
            lr = get_lr(iter_num) if decay_lr else args.lr
            for param_group in optimizer.param_groups:
                param_group["lr"] = lr

            logits = model(X, Y)
            loss = model.last_loss
            scaler.scale(loss).backward()
            train_loss.append(loss.item())

            predictions = torch.argmax(logits, dim=-1)  # Shape: (batch_size, seq_length)
            correct = (predictions == Y).float()
            train_acc.append(correct.mean().item() * 100)

            if grad_clip != 0.0:
                scaler.unscale_(optimizer)
                torch.nn.utils.clip_grad_norm_(model.parameters(), grad_clip)

            # step the optimizer and scaler if training in fp16
            scaler.step(optimizer)
            scaler.update()
            # flush the gradients as soon as we can, no need for this memory anymore
            optimizer.zero_grad(set_to_none=True)

        if iter_num % log_interval == 0:
            dt = time.time() - t0
            print(f"[Round {round}] {iter_num} | "
                  f"loss_cls {loss.item():.4f} | "
                  f"lr {lr:.6f} | {dt * 1000:.2f}ms")
            t0 = time.time()

        iter_num += 1


    return {'mean_train_acc': np.mean(train_acc), 'mean_train_loss': np.mean(train_loss)}



'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''

import numpy as np
import copy
from . import syntax_compiler, baseline
from nncodec.nnc_core import hls
from nncodec.nnc_core import nnr_model
from nncodec.extensions import deepCABAC
from nncodec.nnc_core.nnr_model import NNRModelAccess

def is_block_possible(block_access, approx_data):
    # disable block if decomposed tensors use different approx_methods
    if block_access.dc_g in approx_data['approx_method']:
        if approx_data['approx_method'][block_access.dc_g] != approx_data['approx_method'][block_access.dc_h]:
            return False

    # disable block if non-weights use codebook
    for par_type, param, _ in block_access.param_generator(approx_data["compressed_parameter_types"]):
        if not par_type.endswith("weight"):
            if len(approx_data['approx_method']) > 0 and approx_data['approx_method'][param] == "codebook":
                return False

    # disable block if parameters have inconsistend dq_flags    
    dq_flag = -1
    for _, param, _ in block_access.param_generator(approx_data["compressed_parameter_types"]):
        if dq_flag == -1:
            dq_flag = approx_data["dq_flag"][param]
        else:
            if approx_data["dq_flag"][param] != dq_flag:
                print( "Disabled DQ for block because of inconsistend dq_flags.")
                return False
    
    # disable block if parameters are of type integer
    for par_type, param, _ in block_access.param_generator(approx_data["compressed_parameter_types"]):
        if approx_data['approx_method'][param] == "skip":
            return False
    
    return True


def ndu_enc_generator(enc_info, model_info, approx_data, base_params=None, put_node_depths=None):
    model_access = NNRModelAccess(model_info)
    for block_or_param in model_access.blocks_and_params():
        block_id = block_or_param.block_id
        if block_id is None:
            param = block_or_param.param
            if enc_info.get("out_of_band_signaling", False):
                ndu_oob = syntax_compiler.compile_ndu_oob(  tensor_dims=model_info["parameter_dimensions"][param],
                                                                cabac_unary_length_minus1=enc_info["cabac_unary_length_minus1"],
                                                                compressed_parameter_types=0,
                                                                decomposition_parameter_dict=None
                                                            )
                tensor_dims = None
            # elif enc_info.get("mps_parent_signalling_enabled_flag", 0) and enc_info.get("general_profile_idc", 0):
            #     ndu_oob = syntax_compiler.compile_ndu_oob(model_info["parameter_dimensions"].get(param))
            #     tensor_dims = None
            else:
                tensor_dims = approx_data["parameters"][param].shape
                ndu_oob = syntax_compiler.compile_ndu_oob()

            ndu = syntax_compiler.compile_ndu(
                param=param,
                approx_data=approx_data,
                enc_info=enc_info,
                model_info=model_info,
                ndu_oob=ndu_oob,
                is_block=False,
                cpt=0,
                block_access=None,
                base_params=base_params,
                put_node_depths=put_node_depths,
                tensor_dims=tensor_dims
            )
            yield ndu, [param]
        else:
            cpt = approx_data["compressed_parameter_types"][block_id]
            if "nnr_pt_block_enabled_flag" in enc_info:
                block_condition = is_block_possible( block_or_param, approx_data) and enc_info["nnr_pt_block_enabled_flag"] == 1
            else:
                block_condition = is_block_possible( block_or_param, approx_data)
            if block_condition:
                decomp_parameters = {"decomposition_rank" : approx_data["decomposition_rank"][block_or_param.block_id], "g_number_of_rows" : approx_data["g_number_of_rows"][block_or_param.block_id]} if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0 else None
                if enc_info.get("out_of_band_signaling", False):
                    ndu_oob = syntax_compiler.compile_ndu_oob(  tensor_dims=model_info["parameter_dimensions"][block_or_param.w],
                                                                cabac_unary_length_minus1=enc_info["cabac_unary_length_minus1"],
                                                                compressed_parameter_types=cpt,
                                                                decomposition_parameter_dict=decomp_parameters
                                                            )
                    tensor_dims = None
                elif enc_info.get("mps_parent_signalling_enabled_flag", 0) and enc_info.get("general_profile_idc", 0):
                    ndu_oob = syntax_compiler.compile_ndu_oob(model_info["parameter_dimensions"][block_or_param.w])
                    tensor_dims = None
                else:
                    tensor_dims = model_info["parameter_dimensions"][block_or_param.w]
                    ndu_oob = syntax_compiler.compile_ndu_oob()
                ndu = syntax_compiler.compile_ndu(
                    param=None,
                    approx_data=approx_data,
                    enc_info=enc_info,
                    model_info=model_info,
                    ndu_oob=ndu_oob,
                    is_block=True,
                    cpt=cpt,
                    block_access=block_or_param,
                    base_params=base_params,
                    put_node_depths=put_node_depths,
                    tensor_dims=tensor_dims
                )
                yield ndu, [x for _, x, _ in block_or_param.param_generator(approx_data["compressed_parameter_types"])]
            else:
                for _, param, dims in block_or_param.param_generator(approx_data["compressed_parameter_types"]):
                    if enc_info.get("out_of_band_signaling", False):
                        ndu_oob = syntax_compiler.compile_ndu_oob(  tensor_dims=dims,
                                                                    cabac_unary_length_minus1=enc_info["cabac_unary_length_minus1"],
                                                                    compressed_parameter_types=cpt,
                                                                    decomposition_parameter_dict=decomp_parameters
                                                                )
                        tensor_dims = None
                    elif enc_info.get("mps_parent_signalling_enabled_flag", 0) and enc_info.get("general_profile_idc", 0):
                        ndu_oob = syntax_compiler.compile_ndu_oob(dims)
                        tensor_dims = None
                    else:
                        tensor_dims = dims
                        ndu_oob = syntax_compiler.compile_ndu_oob()
                    ndu = syntax_compiler.compile_ndu(
                        param=param,
                        approx_data=approx_data,
                        enc_info=enc_info,
                        model_info=model_info,
                        ndu_oob=ndu_oob,
                        is_block=False,
                        cpt=cpt,
                        block_access=block_or_param,
                        base_params=base_params,
                        put_node_depths=put_node_depths,
                        tensor_dims=tensor_dims
                    )
                    yield ndu, [param]


def __get_topology_elem_id_order(compressed_parameter_types): ## the FDIS specifies a different order for the tensors and the topology_elem ids within a block. So we have to map them correctly!
    id_list = []

    if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_LS:
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC:
            id_list.append(2)
        else:
            id_list.append(1)
    if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_BI:
        index = 1
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC:
            index += 1
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_LS:
            index += 1
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_BN:
            index += 4
        id_list.append(index)
    if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_BN:
        index = 1
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC:
            index += 1
        if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_LS:
            index += 1
        id_list.append(index)
        id_list.append(index+1)
        id_list.append(index+2)
        id_list.append(index+3)

    id_list.append(0)
    if compressed_parameter_types & hls.BlockParameterTypes.NNR_CPT_DC:
        id_list.append(1)

    return id_list

    
def encode(enc_info, model_info, approx_data, approx_param_base=None, update_base_param=False, tool_if=None):
    ndu_start = syntax_compiler.compile_start_unit(enc_info.get("general_profile_idc", 0))
    bs = hls.encode_nnr_unit_with_size_dummy(ndu_start)
    bs, _ = hls.update_nnr_unit_size(bs)
    mps = syntax_compiler.compile_mps(approx_data,
                                       "topology_storage_format" in model_info,
                                       enc_info.get("general_profile_idc", 0),
                                       enc_info.get("mps_parent_signalling_enabled_flag", 0),
                                       enc_info.get("pre_signalling", 0)
                                      )
    lps = None
    bs_mps = hls.encode_nnr_unit_with_size_dummy( mps )
    bs_mps, _ = hls.update_nnr_unit_size( bs_mps )
    bs.extend(bs_mps)


    if model_info["topology_storage_format"] is not None:
        tpl = syntax_compiler.compile_tpl( model_info )
        bs_tpl = hls.encode_nnr_unit_with_size_dummy( tpl )
        bs_tpl, _ = hls.update_nnr_unit_size( bs_tpl )
        bs.extend(bs_tpl)

    oob_dict = {}

    for ndu, params in ndu_enc_generator(enc_info, model_info, approx_data, approx_param_base if approx_param_base else None, approx_param_base["put_node_depth"] if approx_param_base else None):
        if enc_info.get("general_profile_idc",0) == 0 or not tool_if or not tool_if.hdsp_enabled:
            encoder = deepCABAC.Encoder()
        num_coded_params = 0

        if mps.get("general_profile_idc", 0) and mps.get("mps_parent_signalling_enabled_flag", 0):
            skipped_ndu = True
            for param in params:
                if np.any( approx_data['parameters'][param] ):
                    skipped_ndu = False
                    break
        else:
            skipped_ndu = False

        if not skipped_ndu:
            if enc_info.get("general_profile_idc",0) == 0 or not tool_if or not tool_if.hdsp_enabled:
                for param in params:
                    if ndu["input_parameters_present_flag"] == 0:
                        oob_dict[param] = {"compressed_parameter_types" : ndu["compressed_parameter_types"], "tensor_dimensions" : ndu["tensor_dimensions"], "count_tensor_dimensions" : ndu["count_tensor_dimensions"], "cabac_unary_length_minus1" : ndu["cabac_unary_length_minus1"]}
                        if "decomposition_rank" in ndu and "g_number_of_rows" in ndu:
                            oob_dict[param].update({"decomposition_rank" : ndu["decomposition_rank"], "g_number_of_rows" : ndu["g_number_of_rows"]})
                    else:
                        assert "compressed_parameter_types" in ndu, "compressed_parameter_types must be specified within NDU!"
                        if ndu["compressed_parameter_types"] & hls.BlockParameterTypes.NNR_CPT_DC != 0:
                            assert "decomposition_rank" in ndu, "decomposition_rank must be specified within NDU!"
                            assert "g_number_of_rows" in ndu, "g_number_of_rows must be specified within NDU!"
                        if ndu["tensor_dimensions_flag"] == 0:
                            if param not in oob_dict:
                                oob_dict[param] = {}
                            oob_dict[param].update({"tensor_dimensions" : ndu["tensor_dimensions"], "count_tensor_dimensions" : ndu["count_tensor_dimensions"]})
                        if ndu["cabac_unary_length_flag"] == 0:
                            if param not in oob_dict:
                                oob_dict[param] = {}
                            oob_dict[param].update({ "cabac_unary_length_minus1" : enc_info["cabac_unary_length_minus1"] })
                    if param in approx_data['approx_method']:
                        baseline.encode(
                            encoder,
                            approx_data,
                            approx_param_base if approx_param_base else None,
                            param,
                            ndu,
                            mps,
                            enc_info.get('general_profile_idc', 0),
                            enc_info['param_opt_flag'],
                            enc_info.get( 'row_skip_enabled_flag', 0 ),
                            [],
                            0,
                            lps)
                        num_coded_params += 1

                bs_par = bytearray( encoder.finish().tobytes() )
            else:
                num_modes = tool_if.get_num_modes(params)
                encoders = [deepCABAC.Encoder() for x in range(num_modes)]
                for mode_idx, cur_encoder in enumerate(encoders):
                    for param in params:
                        if param in approx_data['approx_method']:
                            baseline.encode(
                                cur_encoder,
                                approx_data,
                                approx_param_base if approx_param_base else None,
                                param,
                                ndu,
                                mps,
                                enc_info.get('general_profile_idc', 0),
                                enc_info['param_opt_flag'],
                                enc_info.get( 'row_skip_enabled_flag', 0 ),
                                tool_if,
                                mode_idx,
                                lps
                                )
                            if mode_idx == (num_modes-1):
                               num_coded_params += 1

                bit_streams = [ bytearray( cur_enc.finish().tobytes()) for cur_enc in encoders ]

                if num_modes > 1:
                    bs_par = tool_if.get_best_bit_stream_and_set_mode( params, bit_streams )
                else:
                    bs_par           = bit_streams[0]


            decoder = deepCABAC.Decoder()
            decoder.setStream( bs_par )
            approx_data_ep = copy.deepcopy( approx_data )

            epList = np.array([], dtype=np.uint64)

            for param in params:
                if param in approx_data_ep['approx_method']:
                    approx_data_ep["parameters"][param] =  np.zeros_like(approx_data["parameters"][param], dtype=np.int32)
                    epListPart = baseline.decodeAndCreateEPs(
                        decoder,
                        approx_data_ep,
                        approx_param_base if approx_param_base else None,
                        param,
                        ndu,
                        mps,
                        ndu_start,
                        tool_if,
                        lps
                        )
                    if epListPart.size > 0:
                        epList = np.concatenate([epList, epListPart])


            if approx_param_base is not None and update_base_param:
                for param in params:
                    if param not in approx_param_base["put_node_depth"]:
                        approx_param_base["put_node_depth"][
                            param] = 1  # The first node is already the first child node as the base model is the root node
                    else:
                        approx_param_base["put_node_depth"][param] += 1
                    if not approx_param_base["device_id"]:
                        approx_param_base["device_id"] = ndu["device_id"]
                    else:
                        assert approx_param_base["device_id"] == ndu["device_id"], "Unexpected device_id!"
                    if param not in approx_param_base["parameter_id"]:
                        approx_param_base["parameter_id"][param] = ndu["parameter_id"]
                    else:
                        assert approx_param_base["parameter_id"][param] == ndu["parameter_id"], "Unexpected parameter_id!"

                    if enc_info["temporal_context_modeling_flag"]:
                        approx_param_base["parameters"][param] = copy.deepcopy(approx_data["parameters"][param])


            ndu = syntax_compiler.compile_ndu_eps( ndu, epList )

            bs_ndu = hls.encode_nnr_unit_with_size_dummy(ndu)
            if num_coded_params > 0:
                bs_ndu.extend( bs_par )
            bs_ndu, _ = hls.update_nnr_unit_size(bs_ndu)
            bs.extend( bs_ndu )
    return bs, oob_dict


def __decode_nnr_start_unit(nnr_gen, ndu_start, hls_stats = {}):
    next(nnr_gen) # resume to final yield
    bytes_start = next(nnr_gen)
    assert bytes_start == ndu_start["nnr_unit_size"], "nnr_unit_size doesn't match the number of decoded bytes."
    hls_stats["start_bytes"] = bytes_start
    return bytes_start

def __decode_nnr_mps_unit(nnr_gen, reader, ndu_mps, ndu_start, hls_stats = {}):
    next(nnr_gen) # final yield of nnr_unit_size_and_header
    ndu_mps.update({ "general_profile_idc": ndu_start["general_profile_idc"]})
    hls.decode_nnr_unit_payload(reader, ndu_mps)
    bytes_ndu = reader.getNumBytesTouched()
    assert bytes_ndu == ndu_mps["nnr_unit_size"], "nnr_unit_size doesn't match the number of decoded bytes."
    hls_stats["mps_bytes"] = bytes_ndu
    return bytes_ndu

def __decode_nnr_lps_unit(nnr_gen, reader, ndu_lps, hls_stats = {}):
    next(nnr_gen) # final yield of nnr_unit_size_and_header
    hls.decode_nnr_unit_payload(reader, ndu_lps)
    bytes_ndu = reader.getNumBytesTouched()
    assert bytes_ndu == ndu_lps["nnr_unit_size"], "nnr_unit_size doesn't match the number of decoded bytes."
    hls_stats["lps_bytes"] = bytes_ndu
    return bytes_ndu

def __decode_nnr_tpl_unit(nnr_gen, reader, ndu_tpl, mps, model_info, hls_stats={}):
    ndu_tpl.update({ "mps_topology_indexed_reference_flag": mps["mps_topology_indexed_reference_flag"]})
    next(nnr_gen)  # final yield of nnr_unit_size_and_header
    hls.decode_nnr_unit_payload(reader, ndu_tpl)
    bytes_ndu = reader.getNumBytesTouched()
    assert bytes_ndu == ndu_tpl["nnr_unit_size"], "nnr_unit_size doesn't match the number of decoded bytes."
    hls_stats["bytes_tpl"] = bytes_ndu
    model_info["topology_storage_format"] = ndu_tpl["topology_storage_format"]
    if ndu_tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_UNREC:
        pass
    elif ndu_tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_NNEF:
        raise NotImplementedError("NNEF Topology Storage Format not yet implemented!")
    elif ndu_tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_ONNX:
        raise NotImplementedError("ONNX Topology Storage Format not yet implemented!")
    elif ndu_tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_PYT:
        pass
    elif ndu_tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_TEF:
        pass
    elif ndu_tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_PRUN:
        raise NotImplementedError("PRUN Topology Storage Format not yet implemented!")
    elif ndu_tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST:
        raise NotImplementedError("REFLIST Topology Storage Format not yet implemented!")
    return bytes_ndu

def __decode_nnr_qnt_unit():
    pass

def __decode_nnr_ndu_unit(nnr_gen, reader, bitstream, ndu, mps, lps, tpl, ndu_start, model_info, approx_data, bytes_read,
                          decoded_dc_tensorG, tool_if, hls_stats={}, set_model_info=True, oob_dict=None,
                          approx_param_base=None, update_base_param=False):
    block_id = None
    parameter_index = len(model_info["parameter_index"].keys())
    add_block_id_to_model_info = False

    ndu.update({"mps_topology_indexed_reference_flag": mps["mps_topology_indexed_reference_flag"]})  # Propagate mps_topology_indexed_reference_flag to the ndu so it is available when decoding
    ndu.update({"mps_parent_signalling_enabled_flag": mps.get("mps_parent_signalling_enabled_flag", 0)})  # Propagate mps_parent_signalling_enabled_flag to the ndu so it is available when decoding
    ndu.update( {"general_profile_idc" : ndu_start.get("general_profile_idc", 0)} )

    next(nnr_gen)  # continue after decoding nnr unit type and stop after input_parameters_present_flag

    ndu_oob = syntax_compiler.compile_ndu_oob_from_dict(oob_dict, ndu, model_info)
    ndu.update(ndu_oob)

    if ndu["input_parameters_present_flag"] == 0:
        ndu.update({"compressed_parameter_types": 0})  ##It seems the FDIS doesn't say anything about compressed_parameter_types in this case! Setting it to zero should be fine!
    if not "cabac_unary_length_minus1" in ndu:
        ndu.update({"cabac_unary_length_minus1": 9})  ##If it is not decoded from the bitstream use 10 as default

    cpt = ndu["compressed_parameter_types"]
    if decoded_dc_tensorG:
        assert cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0, "Preceding NDU contained a decomposed tensor G! This NDU must contain tensor H!"
    if ndu["nnr_compressed_data_unit_payload_type"] != hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK:
        parType = "unspecified"
        if ndu["nnr_multiple_topology_elements_present_flag"]:
            raise NotImplementedError("Tensor Split not yet implemented!")
        else:
            if not ndu["mps_topology_indexed_reference_flag"]:
                param = ndu["topology_elem_id"]
            else:
                assert tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST, "TPL Unit must be of type NNR_TPL_REFLIST!"
                topology_elem_id_index = ndu["topology_elem_id_index"]
                param = model_info["topology_elem_id_list"][topology_elem_id_index]
        # if ndu["input_parameters_present_flag"] == 0:
        #     ndu_update = {"tensor_dimensions" : model_info["parameter_dimensions"][param], "count_tensor_dimensions" : len(model_info["parameter_dimensions"][param])}#syntax_compiler.compile_ndu_oob(model_info["parameter_dimensions"][param])
        #     ndu.update(ndu_update)
        # elif ndu["tensor_dimensions_flag"] == 0:
        #     if param in model_info["parameter_index"]:
        #         ndu_update = {"count_tensor_dimensions": len(model_info["parameter_dimensions"][param]),
        #                       "tensor_dimensions": model_info["parameter_dimensions"][param]}
        #     # INCTM (V2): if NNR_PT_BLOCK-parameters are coded separately, i.e., is_block_possible() returns False
        #     elif param.endswith(".weight.ls") or param.endswith(".weight.bias"):
        #         ndu_update = {"count_tensor_dimensions": 1,
        #                       "tensor_dimensions": [model_info["parameter_dimensions"][param.replace(
        #                           f'.{param.split(".")[-1]}', '')][0]]}
        #     else:
        #         assert 0, "Block parameter not implemented."
        #     ndu.update(ndu_update)
        next(nnr_gen)  # continue decoding of the nnr unit size and header after input_parameters_present_flag
        if ndu["partial_data_counter_present_flag"] == 1:
            assert 0, "partial_data_counter not yet implemented!"
        hls.decode_nnr_unit_payload(reader, ndu)
        bytes_ndu = reader.getNumBytesTouched()

        if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
            add_block_id_to_model_info = param[:-2] not in model_info["block_identifier"]
            parType = "weight"
            if not decoded_dc_tensorG and param.endswith("_G"):  ##param.endwith is optional!
                ndu["_decomposed_tensor_type"] = "G"
                decoded_dc_tensorG = True
            elif decoded_dc_tensorG and param.endswith("_H"):
                ndu["_decomposed_tensor_type"] = "H"
                decoded_dc_tensorG = False

        params = [(parType, param, ndu["tensor_dimensions"])]
    else:
        assert ndu["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK, "Payload Type must be NNR_PT_BLOCK"
        assert ndu["nnr_multiple_topology_elements_present_flag"], "nnr_multiple_topology_elements_present_flag must be equal to one for NNR_PT_BLOCK"

        block_id = None
        params = []
        id_list_counter = 0
        id_index_list = __get_topology_elem_id_order(cpt)

        ##Get tensor_dimensions and count_tensor_dimensions if neccessary (If they are not transmitted in the bitstream)
        if not ndu["mps_topology_indexed_reference_flag"]:
            weight_param = ndu["topology_elem_id_list"][id_index_list[-1]]
        else:
            assert tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST, "TPL Unit must be of type NNR_TPL_REFLIST!"
            raise NotImplementedError("REFLIST Topology storage format not yet implemented!")
        if weight_param.endswith("_G") or weight_param.endswith("_H"):
            weight_param = weight_param[:-2]

        add_block_id_to_model_info = weight_param not in model_info["block_identifier"]

        if not ndu_oob and (ndu["input_parameters_present_flag"] == 0 or ndu.get("tensor_dimensions_flag") == 0):
            assert weight_param in model_info["parameter_dimensions"], "The tensor dimensions must be present in the model_info, if they are not transmitted in the bitstream!"
            if ndu["input_parameters_present_flag"] == 0:
                ndu_update = {"tensor_dimensions" : model_info["parameter_dimensions"][weight_param], "count_tensor_dimensions" : len(model_info["parameter_dimensions"][weight_param])} ##TODO HAASE: MAYBE THIS WON'T WORK AS INTENDED
                ndu.update(ndu_update)
                #ndu_update = syntax_compiler.compile_ndu_oob(
                #    model_info["parameter_dimensions"][weight_param])  ##derive param name
                #ndu.update(ndu_update)
            elif ndu["tensor_dimensions_flag"] == 0:
                ndu_update = {"count_tensor_dimensions": len(model_info["parameter_dimensions"][weight_param]),
                              "tensor_dimensions": model_info["parameter_dimensions"][weight_param]}
                ndu.update(ndu_update)

        next(nnr_gen)  # continue decoding of the nnr unit size and header
        if ndu["partial_data_counter_present_flag"] == 1:
            assert 0, "partial_data_counter not yet implemented!"
        hls.decode_nnr_unit_payload(reader, ndu)
        bytes_ndu = reader.getNumBytesTouched()

        if cpt & hls.BlockParameterTypes.NNR_CPT_LS != 0:
            parType = "weight.ls"
            if not ndu["mps_topology_indexed_reference_flag"]:
                param = ndu["topology_elem_id_list"][id_index_list[id_list_counter]]
            else:
                assert tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST, "TPL Unit must be of type NNR_TPL_REFLIST!"
                topology_elem_id_list_index = ndu["topology_elem_id_index_list"][id_index_list[id_list_counter]]
                param = model_info["topology_elem_id_list"][topology_elem_id_list_index]
            id_list_counter += 1
            dims = [ndu["tensor_dimensions"][0]]
            params.append( (parType, param, dims) )

        if cpt & hls.BlockParameterTypes.NNR_CPT_BI != 0:
            parType = "bias"
            if not ndu["mps_topology_indexed_reference_flag"]:
                param = ndu["topology_elem_id_list"][id_index_list[id_list_counter]]
            else:
                assert tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST, "TPL Unit must be of type NNR_TPL_REFLIST!"
                topology_elem_id_list_index = ndu["topology_elem_id_index_list"][id_index_list[id_list_counter]]
                param = model_info["topology_elem_id_list"][topology_elem_id_list_index]
            id_list_counter += 1
            dims = [ndu["tensor_dimensions"][0]]
            params.append( (parType, param, dims) )

        if cpt & hls.BlockParameterTypes.NNR_CPT_BN != 0:
            parType = "bn."
            dims = [ndu["tensor_dimensions"][0]]
            for bn_param in ["beta", "gamma", "mean", "var"]:
                if not ndu["mps_topology_indexed_reference_flag"]:
                    param = ndu["topology_elem_id_list"][id_index_list[id_list_counter]]
                else:
                    assert tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST, "TPL Unit must be of type NNR_TPL_REFLIST!"
                    topology_elem_id_list_index = ndu["topology_elem_id_index_list"][id_index_list[id_list_counter]]
                    param = model_info["topology_elem_id_list"][topology_elem_id_list_index]
                id_list_counter += 1
                params.append((parType + bn_param, param, dims))

        if cpt & hls.BlockParameterTypes.NNR_CPT_DC == 0:
            parType = "weight"
            if not ndu["mps_topology_indexed_reference_flag"]:
                param = ndu["topology_elem_id_list"][id_index_list[id_list_counter]]
            else:
                assert tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST, "TPL Unit must be of type NNR_TPL_REFLIST!"
                topology_elem_id_list_index = ndu["topology_elem_id_index_list"][id_index_list[id_list_counter]]
                param = model_info["topology_elem_id_list"][topology_elem_id_list_index]
            id_list_counter += 1
            dims = ndu["tensor_dimensions"]
            params.append((parType, param, dims))

        if cpt & hls.BlockParameterTypes.NNR_CPT_DC != 0:
            parType = "weight"
            for dc_params in ["_G", "_H"]:
                if not ndu["mps_topology_indexed_reference_flag"]:
                    param = ndu["topology_elem_id_list"][id_index_list[id_list_counter]]
                else:
                    assert tpl["topology_storage_format"] == nnr_model.TopologyStorageFormat.NNR_TPL_REFLIST, "TPL Unit must be of type NNR_TPL_REFLIST!"
                    topology_elem_id_list_index = ndu["topology_elem_id_index_list"][id_index_list[id_list_counter]]
                    param = model_info["topology_elem_id_list"][topology_elem_id_list_index]
                id_list_counter += 1
                dims = ndu["tensor_dimensions"]
                params.append((parType, param, dims))

        assert id_list_counter == ndu["count_topology_elements_minus2"] + 2, "Number of decoded topology elements does not match count_topology_elements_minus2 + 2!"

    for par_type, param, dims in params:
        if param.endswith("_G") or param.endswith("_H"):
            param = param[:-2]
        if param in model_info["block_identifier"]:
            block_id = model_info["block_identifier"][param]
        elif add_block_id_to_model_info:
            block_id = "Block_" + weight_param
            model_info["block_identifier"][param] = block_id
        if param not in model_info["parameter_dimensions"] and set_model_info:
            model_info["parameter_dimensions"][param] = dims
        if param not in model_info["parameter_type"] and set_model_info:
            model_info["parameter_type"][param] = par_type
        if param not in model_info["parameter_index"] and set_model_info:
            model_info["parameter_index"][param] = parameter_index
        parameter_index += 1

    if block_id:
        approx_data["compressed_parameter_types"][block_id] = cpt

    hls_stats["ndu_bytes"].append(bytes_ndu)
    if block_id is not None:
        assert approx_data["compressed_parameter_types"][block_id] == cpt
    decoder = deepCABAC.Decoder()
    decoder_initialized = False
    for par_type, param, dims in params:
        if ndu["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_RAW_FLOAT:
            assert param not in approx_data["approx_method"]
            approx_data["parameters"][param] = ndu["raw_float32_parameter"]
            hls_stats["ndu_bytes"][-1] -= 4 * ndu["raw_float32_parameter"].size
        else:
            if (
                (ndu["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
                (ndu["compressed_parameter_types"] & hls.BlockParameterTypes.NNR_CPT_DC != 0) and
                (ndu["codebook_present_flag"] == 1)
            ):
                if param.endswith("_G"):
                    approx_data["approx_method"][param]         = 'codebook'
                    approx_data["codebooks"][param]             = ndu["codebook__"]
                    approx_data["codebook_zero_offsets"][param] = ndu["CbZeroOffset__"]
                    approx_data["codebooks_egk"][param]         = ndu["codebook_egk__"]
                elif param.endswith("_H"):
                    approx_data["approx_method"][param]         = 'codebook'
                    approx_data["codebooks"][param]             = ndu["codebook__dc"]
                    approx_data["codebook_zero_offsets"][param] = ndu["CbZeroOffset__dc"]
                    approx_data["codebooks_egk"][param]         = ndu["codebook_egk__dc"]

            elif (
                  (ndu["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
                  (ndu["codebook_present_flag"] == 1) and
                  (par_type.endswith("weight"))
            ):
                approx_data["approx_method"][param]         = 'codebook'
                approx_data["codebooks"][param]             = ndu["codebook__"]
                approx_data["codebook_zero_offsets"][param] = ndu["CbZeroOffset__"]
                approx_data["codebooks_egk"][param]         = ndu["codebook_egk__"]
            elif(
                 (ndu["nnr_compressed_data_unit_payload_type"] != hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
                 (ndu.get("codebook_present_flag") == 1)
            ):
                approx_data["approx_method"][param]         = 'codebook'
                approx_data["codebooks"][param]             = ndu["codebook__"]
                approx_data["codebook_zero_offsets"][param] = ndu["CbZeroOffset__"]
                approx_data["codebooks_egk"][param]         = ndu["codebook_egk__"]
            elif ndu["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_INT :
                approx_data["approx_method"][param] = 'skip'
            else:
                approx_data["approx_method"][param] = 'uniform'


            if ndu["count_tensor_dimensions"] > 1:
                approx_data["scan_order"][param] = ndu['scan_order']
                if ndu['scan_order'] > 0:
                    tensorDimensions   = dims
                    blockDim           = 4 << ndu['scan_order']

                    if ndu["compressed_parameter_types"] & hls.BlockParameterTypes.NNR_CPT_DC != 0:
                        hNumberOfColumns  = np.int32(np.prod( tensorDimensions )/ndu["g_number_of_rows"])
                        tensorDimensionsG = [ndu["g_number_of_rows"], ndu["decomposition_rank"]]
                        tensorDimensionsH = [ndu["decomposition_rank"], hNumberOfColumns]

                    if (
                        (ndu["nnr_compressed_data_unit_payload_type"] != hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
                        (ndu["compressed_parameter_types"] & hls.BlockParameterTypes.NNR_CPT_DC != 0)
                    ):
                        if ndu["_decomposed_tensor_type"] == "G":
                            tensorDimensions = tensorDimensionsG
                        elif ndu["_decomposed_tensor_type"] == "H":
                            tensorDimensions = tensorDimensionsH

                    numBlockRowsMinus1 = ((tensorDimensions[0]+blockDim-1) >> (2+ndu['scan_order'])) - 1
                    entryPoints = ndu["cabac_entry_point_list"][0:numBlockRowsMinus1]

                    if (
                        (ndu["nnr_compressed_data_unit_payload_type"] == hls.CompressedDataUnitPayloadType.NNR_PT_BLOCK) and
                        (ndu["compressed_parameter_types"] & hls.BlockParameterTypes.NNR_CPT_DC != 0)
                    ):
                        numBlockRowsMinus1G  = ((tensorDimensionsG[0]+blockDim-1) >> (2+ndu["scan_order"])) - 1
                        numBlockRowsMinus1H  = ((tensorDimensionsH[0]+blockDim-1) >> (2+ndu["scan_order"])) - 1

                        if param.endswith("_G"):
                            entryPoints =  ndu["cabac_entry_point_list"][0:numBlockRowsMinus1G]
                        elif param.endswith("_H"):
                            entryPoints =  ndu["cabac_entry_point_list"][numBlockRowsMinus1G:(numBlockRowsMinus1G+numBlockRowsMinus1H)]

                    decoder.setEntryPoints( entryPoints )


            tensorDimensions = dims

            if ndu["compressed_parameter_types"] & hls.BlockParameterTypes.NNR_CPT_DC != 0:
                hNumberOfColumns  = np.int32(np.prod( tensorDimensions )/ndu["g_number_of_rows"])
                tensorDimensionsG = [ndu["g_number_of_rows"], ndu["decomposition_rank"]]
                tensorDimensionsH = [ndu["decomposition_rank"], hNumberOfColumns]

            if param.endswith("_G"):
                dims = tensorDimensionsG
            elif param.endswith("_H"):
                dims = tensorDimensionsH

            approx_data["parameters"][param] = np.zeros(dims, dtype=np.int32)
            if bytes_ndu != 0:  # Decode only if it is not a skipped ndu
                if not decoder_initialized:
                    decoder.setStream(bitstream[bytes_read + bytes_ndu:])
                    decoder_initialized = True
                baseline.decode(
                    decoder,
                    approx_data,
                    approx_param_base if approx_param_base else None,
                    param,
                    ndu,
                    mps,
                    ndu_start,
                    tool_if,
                    lps
                )
                if approx_param_base is not None and update_base_param:
                    if param not in approx_param_base["put_node_depth"]:
                        approx_param_base["put_node_depth"][
                            param] = 1  # The first node is already the first child node as the base model is the root node
                    else:
                        approx_param_base["put_node_depth"][param] += 1
                    if not approx_param_base["device_id"]:
                        approx_param_base["device_id"] = ndu["device_id"]
                    else:
                        assert approx_param_base["device_id"] == ndu["device_id"], "Unexpected device_id!"
                    if param not in approx_param_base["parameter_id"]:
                        approx_param_base["parameter_id"][param] = ndu["parameter_id"]
                    else:
                        assert approx_param_base["parameter_id"][param] == ndu["parameter_id"], "Unexpected parameter_id!"
                    approx_param_base["parameters"][param] = copy.deepcopy(approx_data["parameters"][param])
            else:
                approx_data["qp"][param] = 0
                approx_data["dq_flag"][param] = 0


        if lps is not None:
            model_info["performance_map_flags"]["lps_sparsification_flag"][param]                = lps["lps_sparsification_flag"]
            model_info["performance_map_flags"]["lps_pruning_flag"][param]                       = lps["lps_pruning_flag"]
            model_info["performance_map_flags"]["lps_unification_flag"][param]                   = lps["lps_unification_flag"]
            model_info["performance_map_flags"]["lps_decomposition_performance_map_flag"][param] = lps["lps_decomposition_performance_map_flag"]
        else:
            model_info["performance_map_flags"]["lps_sparsification_flag"][param]                = 0
            model_info["performance_map_flags"]["lps_pruning_flag"][param]                       = 0
            model_info["performance_map_flags"]["lps_unification_flag"][param]                   = 0
            model_info["performance_map_flags"]["lps_decomposition_performance_map_flag"][param] = 0

        if mps is not None:
            model_info["performance_map_flags"]["mps_sparsification_flag"][param]                = mps["mps_sparsification_flag"]
            model_info["performance_map_flags"]["mps_pruning_flag"][param]                       = mps["mps_pruning_flag"]
            model_info["performance_map_flags"]["mps_unification_flag"][param]                   = mps["mps_unification_flag"]
            model_info["performance_map_flags"]["mps_decomposition_performance_map_flag"][param] = mps["mps_decomposition_performance_map_flag"]
        else:
            model_info["performance_map_flags"]["mps_sparsification_flag"][param]                = 0
            model_info["performance_map_flags"]["mps_pruning_flag"][param]                       = 0
            model_info["performance_map_flags"]["mps_unification_flag"][param]                   = 0
            model_info["performance_map_flags"]["mps_decomposition_performance_map_flag"][param] = 0

    if decoder_initialized:
        bytes_ndu += decoder.finish()
    assert bytes_ndu == ndu["nnr_unit_size"], "nnr_unit_size doesn't match the number of decoded bytes."

    return bytes_ndu, decoded_dc_tensorG


def __decode_nnr_unit(reader, bitstream, bytes_read, ndu_start, mps, lps, tpl, model_info, approx_data, nnr_ndu_decoded, decoded_dc_tensorG, set_model_info, tool_if, approx_param_base, update_base_param, oob_dict, hls_stats={}):
    bytes_ndu = 0
    ndu = {}
    g = hls.decode_nnr_unit_size_and_header(reader, ndu)
    next(g) # start decoding of the nnr unit size and header and stop at nnr unit type

    if decoded_dc_tensorG:
        assert ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_NDU, "Preceding NDU contained a decomposed tensor G! This NDU must be of type NNR_NDU!"

    if ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_MPS:
        assert mps is None, "NNR_MPS Unit already decoded! There shall be only one NNR_MPS in the bitstream!"
        assert not nnr_ndu_decoded, "An NNR_MPS shall precede any NNR_NDU, but an NNR_NDU has already been decoded!"
        bytes_ndu = __decode_nnr_mps_unit( g, reader, ndu, ndu_start, hls_stats )
        mps = ndu
        if "mps_qp_density" in mps:
            approx_data.update( { "qp" : {}, "qp_density" : np.int32(mps["mps_qp_density"]), "dq_flag" : {} } )

        if mps["mps_sparsification_flag"] == 1 :
            model_info["performance_maps"]["mps"]["sparsification_performance_map"] = {
                "spm_count_thresholds":     mps["spm_count_thresholds"],
                "sparsification_threshold": mps["sparsification_threshold"],
                "non_zero_ratio":           mps["non_zero_ratio"],
                "spm_nn_accuracy":          mps["spm_nn_accuracy"],
                "spm_count_classes":        mps["spm_count_classes"],
                "spm_class_bitmask":        mps["spm_class_bitmask"],
                "spm_nn_class_accuracy":    mps["spm_nn_class_accuracy"]
            }
        if mps["mps_pruning_flag"] == 1 :
            model_info["performance_maps"]["mps"]["pruning_performance_map"] = {
                "ppm_count_pruning_ratios": mps["ppm_count_pruning_ratios"],
                "pruning_ratio":            mps["pruning_ratio"],
                "ppm_nn_accuracy":          mps["ppm_nn_accuracy"],
                "ppm_count_classes":        mps["ppm_count_classes"],
                "ppm_class_bitmask":        mps["ppm_class_bitmask"],
                "ppm_nn_class_accuracy":    mps["ppm_nn_class_accuracy"]
            }
        if mps["mps_unification_flag"] == 1 :
            model_info["performance_maps"]["mps"]["unification_performance_map"] = {
                "upm_count_thresholds":            mps["upm_count_thresholds"],
                "count_reshaped_tensor_dimension": mps["count_reshaped_tensor_dimension"],
                "reshaped_tensor_dimensions":      mps["reshaped_tensor_dimensions"],
                "count_super_block_dimension":     mps["count_super_block_dimension"],
                "super_block_dimensions":          mps["super_block_dimensions"],
                "count_block_dimension":           mps["count_block_dimension"],
                "block_dimensions":                mps["block_dimensions"],
                "unification_threshold":           mps["unification_threshold"],
                "upm_nn_accuracy":                 mps["upm_nn_accuracy"],
                "upm_count_classes":               mps["upm_count_classes"],
                "upm_class_bitmask":               mps["upm_class_bitmask"],
                "upm_nn_class_accuracy":           mps["upm_nn_class_accuracy"]
            }
        if mps["mps_decomposition_performance_map_flag"] == 1 :
            model_info["performance_maps"]["mps"]["decomposition_performance_map"] = {
                "dpm_count_thresholds":  mps["dpm_count_thresholds"],
                "mse_threshold":         mps["mse_threshold"],
                "dpm_nn_accuracy":       mps["dpm_nn_accuracy"],
                "nn_reduction_ratio":    mps["nn_reduction_ratio"],
                "dpm_count_classes":     mps["dpm_count_classes"],
                "dpm_nn_class_accuracy": mps["dpm_nn_class_accuracy"]
            }

    elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_LPS:
        bytes_ndu = __decode_nnr_lps_unit( g, reader, ndu, hls_stats )
        lps = ndu
        if "lps_qp_density" in lps:
            approx_data.update( { "qp" : {}, "qp_density" : np.int32(lps["lps_qp_density"]), "dq_flag" : {} } )

        if lps["lps_sparsification_flag"] == 1 :
            model_info["performance_maps"]["lps"]["sparsification_performance_map"] = {
                "spm_count_thresholds":     lps["spm_count_thresholds"],
                "sparsification_threshold": lps["sparsification_threshold"],
                "non_zero_ratio":           lps["non_zero_ratio"],
                "spm_nn_accuracy":          lps["spm_nn_accuracy"],
                "spm_count_classes":        lps["spm_count_classes"],
                "spm_class_bitmask":        lps["spm_class_bitmask"],
                "spm_nn_class_accuracy":    lps["spm_nn_class_accuracy"]
            }
        if lps["lps_pruning_flag"] == 1 :
            model_info["performance_maps"]["lps"]["pruning_performance_map"] = {
                "ppm_count_pruning_ratios": lps["ppm_count_pruning_ratios"],
                "pruning_ratio":            lps["pruning_ratio"],
                "ppm_nn_accuracy":          lps["ppm_nn_accuracy"],
                "ppm_count_classes":        lps["ppm_count_classes"],
                "ppm_class_bitmask":        lps["ppm_class_bitmask"],
                "ppm_nn_class_accuracy":    lps["ppm_nn_class_accuracy"]
            }
        if lps["lps_unification_flag"] == 1 :
            model_info["performance_maps"]["lps"]["unification_performance_map"] = {
                "upm_count_thresholds":            lps["upm_count_thresholds"],
                "count_reshaped_tensor_dimension": lps["count_reshaped_tensor_dimension"],
                "reshaped_tensor_dimensions":      lps["reshaped_tensor_dimensions"],
                "count_super_block_dimension":     lps["count_super_block_dimension"],
                "super_block_dimensions":          lps["super_block_dimensions"],
                "count_block_dimension":           lps["count_block_dimension"],
                "block_dimensions":                lps["block_dimensions"],
                "unification_threshold":           lps["unification_threshold"],
                "upm_nn_accuracy":                 lps["upm_nn_accuracy"],
                "upm_count_classes":               lps["upm_count_classes"],
                "upm_class_bitmask":               lps["upm_class_bitmask"],
                "upm_nn_class_accuracy":           lps["upm_nn_class_accuracy"]
            }
        if lps["lps_decomposition_performance_map_flag"] == 1 :
            model_info["performance_maps"]["lps"]["decomposition_performance_map"] = {
                "dpm_count_thresholds":  lps["dpm_count_thresholds"],
                "mse_threshold":         lps["mse_threshold"],
                "dpm_nn_accuracy":       lps["dpm_nn_accuracy"],
                "nn_reduction_ratio":    lps["nn_reduction_ratio"],
                "dpm_count_classes":     lps["dpm_count_classes"],
                "dpm_nn_class_accuracy": lps["dpm_nn_class_accuracy"]
            }

    elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_TPL:
        bytes_ndu = __decode_nnr_tpl_unit( g, reader, ndu, mps, model_info, hls_stats )
        tpl = ndu

    elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_QNT:
        raise NotImplementedError("Decoding of NNR_LPS Units not yet implemented!")

    elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_AGG:
        raise NotImplementedError("Decoding of NNR_AGG Units not yet implemented!")

    elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_NDU:
        bytes_ndu, decoded_dc_tensorG = __decode_nnr_ndu_unit(g, reader, bitstream, ndu, mps, lps, tpl, ndu_start,
                                                              model_info, approx_data, bytes_read, decoded_dc_tensorG,
                                                              tool_if, hls_stats, set_model_info, oob_dict,
                                                              approx_param_base, update_base_param)

    else:
        assert 0, "nnr_unit_type: {} is not specified!".format(ndu["nnr_unit_type"])

    nnr_ndu_decoded = True

    return bytes_ndu, mps, lps, tpl, model_info, approx_data, nnr_ndu_decoded, decoded_dc_tensorG

def decode(bitstream, model_info, oob_dict = None , tool_if=None, hls_stats = {}, approx_param_base=None, update_base_param=False, return_hls=False):
    assert isinstance(bitstream, (bytearray, bytes))

    if not isinstance(bitstream, bytearray):
        bitstream = bytearray(bitstream)

    hls_stats["ndu_bytes"] = []
    approx_data = {
        "approx_method": {},
        "parameters": {},
        "compressed_parameter_types": {},
        "scan_order": {},
        "codebooks": {},
        "codebooks_egk": {},
        "codebook_zero_offsets": {}
    }
    mps = None
    lps = None
    tpl = None
    nnr_ndu_decoded = False
    decoded_dc_tensorG = False ##Only required if DC but not NNR_PT_BLOCK
    set_model_info = False if len(model_info["parameter_type"]) != 0 else True

    ndu_start = {}
    nnr_gen = hls.decode_nnr_unit(bitstream, ndu_start)
    next(nnr_gen) # start decoding and stop at nnr unit type

    assert ndu_start["nnr_unit_type"] == hls.NnrUnitType.NNR_STR, "First nnr unit shall be of type NNR_STR."
    bytes_start = __decode_nnr_start_unit(nnr_gen, ndu_start, hls_stats )
    bytes_read = [bytes_start]

    while( bytes_read[0] < len(bitstream) ): ##Check if there are still remaining NNR Units! ##TODO: Consider having an own function for each case!

        reader = hls.BitReader(bitstream[bytes_read[0]:])

        bytes_ndu, mps, lps, tpl, model_info, approx_data, nnr_ndu_decoded, decoded_dc_tensorG = __decode_nnr_unit( reader,
                                                                                                                    bitstream,
                                                                                                                    bytes_read[0],
                                                                                                                    ndu_start,
                                                                                                                    mps,
                                                                                                                    lps,
                                                                                                                    tpl,
                                                                                                                    model_info,
                                                                                                                    approx_data,
                                                                                                                    nnr_ndu_decoded,
                                                                                                                    decoded_dc_tensorG,
                                                                                                                    set_model_info,
                                                                                                                    tool_if,
                                                                                                                    approx_param_base,
                                                                                                                    update_base_param,
                                                                                                                    oob_dict,
                                                                                                                    hls_stats,
                                                                                                                )

        bytes_read[0] += bytes_ndu

    if return_hls:
        approx_data["dec_hls"] = {"mps": mps, "lps": lps, "tpl": tpl}
    return approx_data


def decode_ndu_unit_header(bitstream, model_info, hls_stats={}):
    assert isinstance(bitstream, (bytearray, bytes))
    if not isinstance(bitstream, bytearray):
        bitstream = bytearray(bitstream)

    hls_stats["ndu_bytes"] = []
    approx_data = {
        "approx_method": {},
        "parameters": {},
        "compressed_parameter_types": {},
        "scan_order": {},
        "codebooks": {},
        "codebooks_egk": {},
        "codebook_zero_offsets": {}
    }
    mps = None
    nnr_ndu_decoded = False
    decoded_dc_tensorG = False  ##Only required if DC but not NNR_PT_BLOCK

    ndu_start = {}
    nnr_gen = hls.decode_nnr_unit(bitstream, ndu_start)
    next(nnr_gen)  # start decoding and stop at nnr unit type

    assert ndu_start["nnr_unit_type"] == hls.NnrUnitType.NNR_STR, "First nnr unit shall be of type NNR_STR."
    bytes_start = __decode_nnr_start_unit(nnr_gen, ndu_start, hls_stats)
    bytes_read = [bytes_start]

    while (bytes_read[0] < len(bitstream)):  ##Check if there are still remaining NNR Units! ##TODO: Consider having an own function for each case!
        bytes_ndu = 0
        reader = hls.BitReader(bitstream[bytes_read[0]:])
        ndu = {}
        g = hls.decode_nnr_unit_size_and_header(reader, ndu)
        next(g)  # start decoding of the nnr unit size and header and stop at nnr unit type
        if decoded_dc_tensorG:
            assert ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_NDU, "Preceding NDU contained a decomposed tensor G! This NDU must be of type NNR_NDU!"
        if ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_MPS:
            assert mps is None, "NNR_MPS Unit already decoded! There shall be only one NNR_MPS in the bitstream!"
            assert not nnr_ndu_decoded, "An NNR_MPS shall precede any NNR_NDU, but an NNR_NDU has already been decoded!"
            bytes_ndu = __decode_nnr_mps_unit(g, reader, ndu, ndu_start, hls_stats)
            mps = ndu
            if "mps_qp_density" in mps:
                approx_data.update({"qp": {}, "qp_density": np.int32(mps["mps_qp_density"]), "dq_flag": {}})
            approx_data.update({"pre_flag_model": mps.get("nnr_pre_flag", 0)})
            if model_info is not None: model_info["performance_maps"] = {}
            if mps["mps_sparsification_flag"] == 1 and model_info is not None:
                model_info["performance_maps"]["sparsification"] = {
                    "count_thresholds": mps.get("spm_count_thresholds"),
                    "non_zero_ratio": mps.get("non_zero_ratio"),
                    "sparsification_threshold": mps.get("sparsification_threshold"),
                    "nn_accuracy": mps.get("spm_nn_accuracy"),
                    "count_classes": mps.get("spm_count_classes"),
                    "class_bitmask": mps.get("spm_class_bitmask"),
                    "nn_class_accuracy": mps.get("spm_nn_class_accuracy")
                }
            if mps["mps_pruning_flag"] == 1 and model_info is not None:
                model_info["performance_maps"]["pruning"] = {
                    "count_pruning_ratios": mps.get("ppm_count_pruning_ratios"),
                    "pruning_ratio": mps.get("pruning_ratio"),
                    "nn_accuracy": mps.get("ppm_nn_accuracy"),
                    "count_classes": mps.get("ppm_count_classes"),
                    "class_bitmask": mps.get("ppm_class_bitmask"),
                    "nn_class_accuracy": mps.get("ppm_nn_class_accuracy")
                }
        elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_LPS:
            raise NotImplementedError("Decoding of NNR_LPS Units not yet implemented!")
        elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_TPL:
            bytes_ndu = __decode_nnr_tpl_unit(g, reader, ndu, mps, model_info, hls_stats)
        elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_QNT:
            raise NotImplementedError("Decoding of NNR_LPS Units not yet implemented!")
        elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_AGG:
            raise NotImplementedError("Decoding of NNR_AGG Units not yet implemented!")
        elif ndu["nnr_unit_type"] == hls.NnrUnitType.NNR_NDU:
            ndu.update({"mps_topology_indexed_reference_flag": mps["mps_topology_indexed_reference_flag"]})  # Propagate mps_topology_indexed_reference_flag to the ndu so it is available when decoding
            ndu.update({"mps_parent_signalling_enabled_flag": mps.get("mps_parent_signalling_enabled_flag",0)})  # Propagate mps_parent_signalling_enabled_flag to the ndu so it is available when decoding
            ndu.update({"general_profile_idc": ndu_start.get("general_profile_idc", 0)})

            next(g)  # continue after decoding nnr unit type and stop after input_parameters_present_flag

            return ndu
        bytes_read[0] += bytes_ndu


'''
The copyright in this software is being made available under the Clear BSD
License, included below. No patent rights, trademark rights and/or 
other Intellectual Property Rights other than the copyrights concerning 
the Software are granted under this license.

The Clear BSD License

Copyright (c) 2019-2025, Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V. & The NNCodec Authors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted (subject to the limitations in the disclaimer below) provided that
the following conditions are met:

     * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

     * Neither the name of the copyright holder nor the names of its
     contributors may be used to endorse or promote products derived from this
     software without specific prior written permission.

NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
'''
import copy
import numpy as np
from nncodec.extensions import deepCABAC
from nncodec.nnc_core.nnr_model import NNRModelAccess, W_TYPES

def approx(approx_info, model_info, approx_data_in, enc_info=None):
    approx_data_out = {k: copy.copy(v) for k, v in approx_data_in.items()} # create copies of dicts in approx_data
    encoder = deepCABAC.Encoder()
    model_access = NNRModelAccess(model_info)
    for block_or_param in model_access.blocks_and_params():
        for par_type, param, _ in block_or_param.param_generator(approx_data_in["compressed_parameter_types"]):
            if (par_type in approx_info["to_approximate"]) and (param not in approx_data_in["approx_method"]):
                # !!! There seems to be a pybind11 issue when using np.zeros_like for "values" that have been transposed.
                # !!! It seems that sometimes, encoder.quantLayer returns only zeros for quantizedValues. Needs further study.
                # !!! For now, using np.zeros instead of np.zeros_like seems to be a workaround.           
                quantizedValues = np.zeros(approx_data_in["parameters"][param].shape, dtype=np.int32)
                encoder.initCtxModels( approx_info["cabac_unary_length_minus1"], 0 )

                enc_qp = approx_info['qp'][param]

                qp = encoder.quantLayer(
                    approx_data_in["parameters"][param],
                    quantizedValues,
                    approx_info['dq_flag'][param],
                    approx_data_out['qp_density'],
                    enc_qp,
                    approx_info["lambda_scale"],
                    approx_info["cabac_unary_length_minus1"],
                    approx_data_in["scan_order"].get(param, 0),
                    enc_info.get("general_profile_idc", 0) if enc_info else 0
                )

                if qp != enc_qp:
                    print("INFO: QP for {} has been clipped from {} to {} to avoid int32_t overflow!".format(param, approx_info['qp'][param],qp))
                    approx_data_out['qp'][param] = qp
                else:
                    approx_data_out['qp'][param] = enc_qp

                approx_data_out['parameters'][param] = quantizedValues
                approx_data_out['approx_method'][param] = 'uniform'
                approx_data_out['dq_flag'][param] = approx_info['dq_flag'][param]

                if "integer_aligned_bitdepth" in approx_info and ".num_batches_tracked" not in param:
                    bw_ap = approx_info["integer_aligned_bitdepth"]
                    if bw_ap > 7 or model_info["parameter_type"][param] in W_TYPES:
                        bw = bw_ap
                    else: # non-weight params at least in 16bit
                        bw = 8
                        print(f"INFO: Since {param} seems to be no weight tensor, it was quantized to {bw} bit instead of {bw_ap} bit.")
                    prm = approx_data_in["parameters"][param]
                    if approx_info["unsigned_integer_support"] and len(prm[prm < 0]) == 0:
                        approx_data_out['parameters'][param] = quantizedValues.clip(0, 2 ** bw - 1)
                    else:
                        approx_data_out['parameters'][param] = quantizedValues.clip(-2 ** (bw - 1), 2 ** (bw - 1) - 1)

    return approx_data_out

def rec(param, approx_data):
    assert approx_data['parameters'][param].dtype == np.int32

    decoder = deepCABAC.Decoder()
    values = approx_data['parameters'][param]

    approx_data["parameters"][param] = np.zeros(values.shape, dtype=np.float32)
    decoder.dequantLayer(approx_data["parameters"][param], values, approx_data["qp_density"], approx_data["qp"][param], approx_data['scan_order'].get(param, 0))

    del approx_data["approx_method"][param]

